'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var fs = _interopDefault(require('fs'));
var mime = _interopDefault(require('mime'));
var path = _interopDefault(require('path'));
var rimraf = _interopDefault(require('rimraf'));
var send = _interopDefault(require('send'));
var isModule$2 = _interopDefault(require('is-module'));
var util = _interopDefault(require('util'));
var chalk = _interopDefault(require('chalk'));
var pirates = _interopDefault(require('pirates'));
var glob = _interopDefault(require('glob'));
var module$1 = _interopDefault(require('module'));
var sucrase = _interopDefault(require('sucrase'));
var url$1 = _interopDefault(require('url'));
var pluginCommonjs = _interopDefault(require('@rollup/plugin-commonjs'));
var pluginJson = _interopDefault(require('@rollup/plugin-json'));
var pluginReplace = _interopDefault(require('@rollup/plugin-replace'));
var pluginNodeResolve = _interopDefault(require('@rollup/plugin-node-resolve'));
var rollup$1 = _interopDefault(require('rollup'));
var debug$a = _interopDefault(require('debug'));
var workerFarm = _interopDefault(require('worker-farm'));
var serverDestroy = _interopDefault(require('server-destroy'));
var fayeWebsocket = _interopDefault(require('faye-websocket'));
var getPort = _interopDefault(require('get-port'));
var http = _interopDefault(require('http'));
var https = _interopDefault(require('https'));
var crypto = _interopDefault(require('crypto'));
var permessageDeflate = _interopDefault(require('permessage-deflate'));
var pathToRegexp$1 = _interopDefault(require('path-to-regexp'));
var zlib = _interopDefault(require('zlib'));
var chokidar = _interopDefault(require('chokidar'));

const JS_MIME_TYPES = {
  'application/javascript': ['js', 'jsx', 'ts', 'tsx']
};
const TESTING = process.env.NODE_ENV === 'test';
// Prevent parallel test runs from reading from same cache
const DIR = `.dvlp${TESTING ? process.getuid() : ''}`;
const VERSION = process.env.DVLP_VERSION;

const bundleDirName = `${path.join(DIR, `bundle-${VERSION}`)}`;
const bundleDir = path.resolve(bundleDirName);
const maxModuleBundlerWorkers =
  (process.env.BUNDLE_WORKERS && parseInt(process.env.BUNDLE_WORKERS, 10)) || 0;
const port = process.env.PORT ? Number(process.env.PORT) : 8080;

mime.define(JS_MIME_TYPES, true);
send.mime.define(JS_MIME_TYPES, true);

const dir = path.resolve(DIR);

// Work around @rollup/plugin-commonjs require.main.filename
if (TESTING || process.env.DVLP_LAUNCHER === 'cmd') {
  const bundleDirExists = fs.existsSync(bundleDir);
  const dirExists = fs.existsSync(dir);
  const rm = dirExists && !bundleDirExists;

  if (rm) {
    const contents = fs.readdirSync(dir).map((item) => path.resolve(dir, item));

    for (const item of contents) {
      // Delete all subdirectories
      if (fs.statSync(item).isDirectory()) {
        rimraf.sync(item);
      }
    }
  }
  if (!dirExists) {
    fs.mkdirSync(dir);
  }
  if (!bundleDirExists) {
    fs.mkdirSync(bundleDir);
  }

  if (TESTING) {
    process.on('exit', () => {
      rimraf.sync(dir);
    });
  }
}

/**
 * @type { Config }
 */
var config = {
  activePort: port,
  bundleDir,
  bundleDirName,
  directories: [],
  // Ordered to trigger transpiling if necessary
  extensionsByType: {
    css: ['.sass', '.scss', '.less', '.styl', '.stylus', '.css'],
    html: [
      '.nunjs',
      '.nunjucks',
      '.hbs',
      '.handlebars',
      '.dust',
      '.html',
      '.htm'
    ],
    js: ['.ts', '.tsx', '.mjs', '.jsx', '.js', '.json']
  },
  latency: 50,
  maxAge: '10m',
  maxModuleBundlerWorkers,
  port,
  rollupConfigPath: path.join(dir, 'rollup.config.js'),
  testing: TESTING,
  typesByExtension: {
    '.css': 'css',
    '.sass': 'css',
    '.scss': 'css',
    '.less': 'css',
    '.styl': 'css',
    '.stylus': 'css',
    '.html': 'html',
    '.htm': 'html',
    '.nunjs': 'html',
    '.nunjucks': 'html',
    '.hbs': 'html',
    '.handlebars': 'html',
    '.dust': 'html',
    '.js': 'js',
    '.mjs': 'js',
    '.json': 'js',
    '.jsx': 'js',
    '.ts': 'js',
    '.tsx': 'js'
  }
};

var commonjsGlobal = typeof globalThis !== 'undefined' ? globalThis : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

function createCommonjsModule(fn, module) {
	return module = { exports: {} }, fn(module, module.exports), module.exports;
}

// @ts-ignore




const HAS_UTIL_IS_PROXY = util.types && util.types.isProxy !== undefined;
const RE_BARE_IMPORT = /^[^./]/;
const RE_INVALID = /[<>:"|?*]/;
const RE_JSON = /.json$/i;
const RE_LOCALHOST = /localhost|127\.0\.0\.1/;
const RE_NODE_MODULES = /node_modules/;
const RE_TYPE_CSS = /text\/css/i;
const RE_TYPE_HTML = /text\/html/i;
const RE_TYPE_JS = /application\/javascript/i;

var is = {
  isAbsoluteFilePath,
  isBareImport,
  isCssFilePath,
  isCssRequest,
  isHtmlFilePath,
  isHtmlRequest,
  isInvalidFilePath,
  isJsFilePath,
  isJsRequest,
  isJsonFilePath,
  isLocalhost,
  isNodeModuleFilePath,
  isModule,
  isModuleBundlerFilePath,
  isProjectFilePath,
  isPromise,
  isProxy,
  isRelativeFilePath,
  isValidFilePath
};

/**
 * Determine if "filePath" is absolute
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isAbsoluteFilePath(filePath) {
  return (
    'string' == typeof filePath &&
    path.isAbsolute(filePath) &&
    // Only absolute if from root
    filePath.startsWith(process.cwd().slice(0, 5))
  );
}

/**
 * Determine if 'id' is referencing a node_module
 *
 * @param { string } id
 * @returns { boolean }
 */
function isBareImport(id) {
  return RE_BARE_IMPORT.test(id);
}

/**
 * Determine if 'filePath' is for a css file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isCssFilePath(filePath) {
  return config.extensionsByType.css.includes(path.extname(filePath));
}

/**
 * Determine if 'req' is for a css file
 *
 * @param { Req } req
 * @returns { boolean }
 */
function isCssRequest(req) {
  return (
    req.type === 'css' ||
    (req.headers.accept && RE_TYPE_CSS.test(req.headers.accept)) ||
    isCssFilePath(req.url)
  );
}

/**
 * Determine if 'filePath' is for an html file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isHtmlFilePath(filePath) {
  return config.extensionsByType.html.includes(path.extname(filePath));
}

/**
 * Determine if 'req' is for an html file
 *
 * @param { Req } req
 * @returns { boolean }
 */
function isHtmlRequest(req) {
  return (
    req.type === 'html' ||
    (req.headers.accept && RE_TYPE_HTML.test(req.headers.accept)) ||
    isHtmlFilePath(req.url)
  );
}

/**
 * Determine if 'filePath' is invalid
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isInvalidFilePath(filePath) {
  return RE_INVALID.test(filePath);
}

/**
 * Determine if 'filePath' is for a js file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isJsFilePath(filePath) {
  return config.extensionsByType.js.includes(path.extname(filePath));
}

/**
 * Determine if 'req' is for a js file
 *
 * @param { Req } req
 * @returns { boolean }
 */
function isJsRequest(req) {
  return (
    req.type === 'js' ||
    // Almost always '*/*'
    (req.headers.accept && RE_TYPE_JS.test(req.headers.accept)) ||
    isJsFilePath(req.url)
  );
}

/**
 * Determine if 'filePath' is for a js file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isJsonFilePath(filePath) {
  return RE_JSON.test(filePath);
}

/**
 * Determine if 'url' is localhost
 *
 * @param { string } url
 * @returns { boolean }
 */
function isLocalhost(url) {
  return RE_LOCALHOST.test(url);
}

/**
 * Determine if 'filePath' is in node_modules
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isNodeModuleFilePath(filePath) {
  const isNodeModule = RE_NODE_MODULES.test(filePath);

  if (!isNodeModule) {
    return false;
  }

  try {
    // Resolve symlinks to determine if really a node_module
    return RE_NODE_MODULES.test(fs.realpathSync(filePath));
  } catch (err) {
    return true;
  }
}

/**
 * Determine if filePath or code is es module
 *
 * @param { string } filePathOrCode
 * @returns { boolean }
 */
function isModule(filePathOrCode) {
  if (isJsFilePath(filePathOrCode)) {
    filePathOrCode = fs.readFileSync(filePathOrCode, 'utf8');
  }
  return isModule$2(filePathOrCode);
}

/**
 * Determine if 'filePath' is for a bundled module file
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isModuleBundlerFilePath(filePath) {
  return filePath.includes(config.bundleDirName);
}

/**
 * Determine if 'obj' is Proxy
 *
 * @param { object } obj
 * @returns { boolean }
 */
function isProxy(obj) {
  if (HAS_UTIL_IS_PROXY) {
    return util.types.isProxy(obj);
  } else {
    return obj instanceof Proxy;
  }
}

/**
 * Determine if 'obj' is Promise instance
 *
 * @param { object } obj
 * @returns { boolean }
 */
function isPromise(obj) {
  if (HAS_UTIL_IS_PROXY) {
    return util.types.isPromise(obj);
  } else {
    return obj instanceof Promise;
  }
}

/**
 * Determine if 'filePath' is in project
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isProjectFilePath(filePath) {
  if (!path.isAbsolute(filePath)) {
    filePath = path.resolve(filePath);
  }

  return filePath.includes(process.cwd());
}

/**
 * Determine if "filePath" is relative
 *
 * @param { string } filePath
 * @returns { boolean }
 */
function isRelativeFilePath(filePath) {
  return 'string' == typeof filePath && filePath.startsWith('.');
}

/**
 * Determine if "filePath" is valid.
 * If relative, resolves against "fromFilePath".
 *
 * @param { string } filePath
 * @param { string } [fromDir]
 * @returns { boolean }
 */
function isValidFilePath(filePath, fromDir = process.cwd()) {
  if (isRelativeFilePath(filePath)) {
    filePath = path.join(fromDir, filePath);
  }

  try {
    const stat = fs.statSync(filePath);
    if (stat.isFile()) {
      return true;
    }
  } catch (err) {
    // Ignore
  }
  return false;
}

const SEG_LENGTH = 80;
const WARN_BARE_IMPORT =
  '⚠️  re-writing bare imports because browsers do not support Node-style import identifiers';
const WARN_MISSING_EXTENSION =
  '⚠️  adding missing extensions because browsers can only resolve valid URL identifiers';
const WARN_NODE_PATH =
  '⚠️  re-writing NODE_PATH import because browsers can only resolve valid URL identifiers';
const WARN_PACKAGE_INDEX =
  '⚠️  adding missing package index.js because browsers do not support Node-style package identifiers';
const WARN_SERVER_TRANSPILE =
  '⚠️  ignoring async Promise returned when executing transpiler on a server file. Be sure to check the "isServer" argument before returning a transpiled value';

const disabled = process.env.DVLP_LAUNCHER !== 'cmd';
let maxBareImport = 1;
let maxMissingExtension = 1;
let maxNodePath = 1;
let maxPackageIndex = 1;
let maxServerTranspile = 1;
let silent = false;

var log = {
  WARN_BARE_IMPORT,
  WARN_MISSING_EXTENSION,
  WARN_NODE_PATH,
  WARN_PACKAGE_INDEX,
  WARN_SERVER_TRANSPILE,
  info,
  noisyInfo,
  warn,
  error,
  fatal,

  /**
   * Set silent state
   *
   * @param { boolean } value
   */
  set silent(value) {
    silent = value;
  }
};

/**
 * Log if not testing/silent
 *
 * @param { string } msg
 */
function info(msg) {
  if (!disabled && !silent) {
    console.log(truncate(' ' + msg));
  }
}

/**
 * Log if not testing, even if silent
 *
 * @param { string } msg
 */
function noisyInfo(msg) {
  if (!disabled) {
    console.log(truncate(' ' + msg));
  }
}

/**
 * Warn if not testing/silent
 *
 * @param { ...unknown } args
 */
function warn(...args) {
  if (!disabled) {
    const [warning] = args;
    let msg = '';

    switch (warning) {
      case WARN_BARE_IMPORT:
        if (maxBareImport-- > 0) {
          msg = `${WARN_BARE_IMPORT}`;
        }
        break;
      case WARN_NODE_PATH:
        if (maxNodePath-- > 0) {
          msg = `${WARN_NODE_PATH}`;
        }
        break;
      case WARN_MISSING_EXTENSION:
        if (maxMissingExtension-- > 0) {
          msg = `${WARN_MISSING_EXTENSION}`;
        }
        break;
      case WARN_PACKAGE_INDEX:
        if (maxPackageIndex-- > 0) {
          msg = `${WARN_PACKAGE_INDEX}`;
        }
        break;
      case WARN_SERVER_TRANSPILE:
        if (maxServerTranspile-- > 0) {
          msg = `${WARN_SERVER_TRANSPILE}`;
        }
        break;
      default:
        msg = args.join(' ');
    }

    if (msg) {
      console.warn(msg);
    }
  }
}

/**
 * Error
 *
 * @param { ...unknown } args
 */
function error(...args) {
  if (!disabled) {
    console.error(chalk.red.inverse(' error '), ...args);
  }
}

/**
 * Fatal error
 *
 * @param { ...unknown } args
 */
function fatal(...args) {
  if (!disabled) {
    console.error(chalk.red.inverse(' fatal error '), ...args);
  }
}

/**
 * Truncate 'string'
 *
 * @param { string } string
 * @returns { string }
 */
function truncate(string) {
  if (string.length > SEG_LENGTH * 1.5 + 3) {
    return string.slice(0, SEG_LENGTH) + '...' + string.slice(-SEG_LENGTH / 2);
  }

  return string;
}

var file = createCommonjsModule(function (module) {

const {
  isAbsoluteFilePath,
  isCssRequest,
  isHtmlRequest,
  isJsRequest,
  isModule,
  isModuleBundlerFilePath,
  isNodeModuleFilePath
} = is;
const {
  warn,
  WARN_MISSING_EXTENSION,
  WARN_PACKAGE_INDEX,
  WARN_SERVER_TRANSPILE
} = log;
const { addHook } = pirates;



// Work around @rollup/plugin-commonjs dynamic require
// @ts-ignore
const loadModule = module$1._load;


const { URL } = url$1;

const FILE_TYPES = ['html', 'js', 'css'];
const IMPORT_EXTS = ['.js', '.mjs'];
const IMPORT_EXTS_TRANSPILER = ['.js', '.jsx', '.mjs', '.ts', '.tsx'];
const RE_GLOB = /[*[{]/;
const RE_TRANSPILER_HANDLES_SERVER = /\(\s?[a-zA-Z]+,\s?[a-zA-Z]+\s?\)/;
const RE_SEPARATOR = /[,;]\s?|\s/g;

/** @type { () => void } */
let revertHook;

module.exports = {
  exists,
  expandPath,
  find,
  getAbsoluteProjectPath,
  getProjectPath,
  getTypeFromPath,
  getTypeFromRequest,
  importModule,
  resolveRealFilePath
};

/**
 * Validate that all file paths exist
 *
 * @param { string | Array<string> } filePaths
 * @returns { void }
 */
function exists(filePaths) {
  if (!Array.isArray(filePaths)) {
    filePaths = [filePaths];
  }

  for (const filePath of filePaths) {
    if (!fs.existsSync(path.resolve(filePath))) {
      throw Error(`path '${filePath}' does not exist`);
    }
  }
}

/**
 * Expand "filePath" into multiple filePaths
 * Handles globs and/or separators
 *
 * @param { string | Array<string> } filePath
 * @returns { Array<string> }
 */
function expandPath(filePath) {
  if (!filePath) {
    return [];
  }

  if (typeof filePath === 'string' && fs.existsSync(path.resolve(filePath))) {
    return [filePath];
  }

  if (Array.isArray(filePath)) {
    return filePath.reduce((/** @type { Array<string> } */ filePaths, fp) => {
      if (fp) {
        // @ts-ignore
        filePaths.push(...expandPath(fp));
      }
      return filePaths;
    }, []);
  }

  RE_SEPARATOR.lastIndex = 0;
  if (RE_SEPARATOR.test(filePath)) {
    filePath = filePath.split(RE_SEPARATOR);
  }
  if (!Array.isArray(filePath)) {
    filePath = [filePath];
  }

  return filePath.reduce((/** @type { Array<string> } */ filePaths, fp) => {
    if (RE_GLOB.test(fp)) {
      filePaths.push(...glob.sync(fp));
    } else {
      filePaths.push(fp);
    }
    return filePaths;
  }, []);
}

/**
 * Find filePath for 'req'
 *
 * @param { Req | string } req
 * @param { object } [options]
 * @param { Array<string> } [options.directories]
 * @param { string } [options.type]
 * @returns { string | undefined }
 */
function find(req, { directories = config.directories, type } = {}) {
  const requestedFilePath = isRequestObject(req)
    ? new URL(req.url, 'http://localhost').pathname
    : req;
  let filePath;

  if (!type) {
    type = isRequestObject(req)
      ? getTypeFromRequest(req)
      : getTypeFromPath(req);
  }

  // Handle bundled js import
  if (isModuleBundlerFilePath(requestedFilePath)) {
    filePath = path.join(config.bundleDir, path.basename(requestedFilePath));
  } else if (isAbsoluteFilePath(requestedFilePath)) {
    filePath = resolveFilePath(requestedFilePath, type);
  } else {
    for (const directory of directories) {
      filePath = resolveFilePath(path.join(directory, requestedFilePath), type);

      if (filePath) {
        break;
      }
    }
  }

  if (!filePath) {
    return;
  }

  if (isRequestObject(req)) {
    req.filePath = filePath;
    req.type = getTypeFromPath(filePath);
  }

  return filePath;
}

/**
 * Import esm/cjs module, transpiling if necessary (via require hook)
 *
 * @param { string } modulePath
 * @param { Transpiler } [transpiler]
 * @returns { object }
 */
function importModule(modulePath, transpiler) {
  if (revertHook !== undefined) {
    revertHook();
  }

  revertHook = addHook(
    (code, filePath) => {
      // Determine if transpiler supports transpiling server modules by checking number of arguments (filePath, isServer) handled
      if (
        transpiler !== undefined &&
        RE_TRANSPILER_HANDLES_SERVER.test(transpiler.toString())
      ) {
        const transpiled = transpiler(filePath, true);

        if (transpiled) {
          // Ignore async
          if (transpiled instanceof Promise) {
            warn(WARN_SERVER_TRANSPILE);
          } else {
            code = transpiled;
          }
        }
      }

      if (!isModule(code)) {
        return code;
      }

      return sucrase.transform(code, {
        transforms: ['imports'],
        filePath
      }).code;
    },
    {
      exts: transpiler ? IMPORT_EXTS_TRANSPILER : IMPORT_EXTS,
      ignoreNodeModules: false
    }
  );
  let mod = loadModule(modulePath, module, false);

  // Return default if only exported key
  if ('default' in mod && Object.keys(mod).length === 1) {
    mod = mod.default;
  }

  return mod;
}

/**
 * Retrieve the project relative path for "filePath"
 *
 * @param { string | Array<string> } filePath
 * @returns { string }
 */
function getProjectPath(filePath) {
  if (Array.isArray(filePath)) {
    filePath = filePath[0];
  }

  const projectFilePath = isAbsoluteFilePath(filePath)
    ? path.relative(process.cwd(), filePath)
    : filePath;

  return projectFilePath.startsWith('/')
    ? projectFilePath.slice(1)
    : projectFilePath;
}

/**
 * Retrieve the absolute path for the project relative path "filePath"
 *
 * @param { string } filePath
 * @returns { string }
 */
function getAbsoluteProjectPath(filePath) {
  return isAbsoluteFilePath(filePath)
    ? filePath
    : path.join(
        process.cwd(),
        filePath.charAt(0) === '/' ? filePath.slice(1) : filePath
      );
}

/**
 * Retrieve resource type
 *
 * @param { Req } req
 * @returns { string }
 */
function getTypeFromRequest(req) {
  if (req.type) {
    return req.type;
  } else if (isHtmlRequest(req)) {
    return 'html';
  } else if (isCssRequest(req)) {
    return 'css';
  } else if (isJsRequest(req)) {
    return 'js';
  } else {
    return '';
  }
}

/**
 * Retrieve generic file type from "filePath" extension
 *
 * @param { string } filePath
 * @returns { string }
 */
function getTypeFromPath(filePath) {
  return config.typesByExtension[path.extname(filePath)];
}

/**
 * Resolve "filePath" of "type"
 * Handles missing extensions and package indexes
 *
 * @param { string } filePath
 * @param { string } type
 * @returns { string }
 */
function resolveFilePath(filePath, type) {
  // prettier-ignore
  filePath = decodeURI(filePath).replace(/(\s)/g, '\$1'); // eslint-disable-line

  try {
    const stat = fs.statSync(filePath);

    if (stat.isFile()) {
      return resolveRealFilePath(filePath);
    }
  } catch (err) {
    // Not found, possibly no extension or package
  }

  if (!type) {
    for (type of FILE_TYPES) {
      const fp = resolveFilePath(filePath, type);
      if (fp) {
        return resolveRealFilePath(fp);
      }
    }
  }

  let fp = resolveFilePathExtension(filePath, config.extensionsByType[type]);
  if (fp) {
    if (!isNodeModuleFilePath(fp)) {
      warn(WARN_MISSING_EXTENSION, getProjectPath(filePath));
    }
    return resolveRealFilePath(fp);
  }

  fp = resolveFilePathExtension(
    path.join(filePath, 'index'),
    config.extensionsByType[type]
  );

  if (fp && type === 'js') {
    if (!isNodeModuleFilePath(fp)) {
      warn(WARN_PACKAGE_INDEX, getProjectPath(filePath));
    }
  }

  return resolveRealFilePath(fp);
}

/**
 * Resolve missing extension for "filePath"
 *
 * @param { string } filePath
 * @param { Array<string> } extensions
 * @returns { string }
 */
function resolveFilePathExtension(filePath, extensions) {
  for (const ext of extensions) {
    const fp = filePath + ext;

    if (fs.existsSync(fp)) {
      return fp;
    }
  }

  return '';
}

/**
 * Resolve real path to "filePath", even if symlinked
 *
 * @param { string } filePath
 * @returns { string }
 */
function resolveRealFilePath(filePath) {
  if (!filePath) {
    return filePath;
  }

  try {
    return fs.realpathSync(filePath);
  } catch (err) {
    return filePath;
  }
}

/**
 * Determine whether "req" is a request object
 *
 * @param { unknown } req
 * @returns { req is Req }
 */
function isRequestObject(req) {
  return typeof req !== 'string';
}
});
var file_1 = file.exists;
var file_2 = file.expandPath;
var file_3 = file.find;
var file_4 = file.getAbsoluteProjectPath;
var file_5 = file.getProjectPath;
var file_6 = file.getTypeFromPath;
var file_7 = file.getTypeFromRequest;
var file_8 = file.importModule;
var file_9 = file.resolveRealFilePath;

const { rollup } = rollup$1;

/** @typedef { import("rollup").InputOptions } InputOptions */
/** @typedef { import("rollup").OutputOptions } OutputOptions */
/** @typedef { import("rollup").RollupOptions } RollupOptions */

/** @type { InputOptions } */
const ROLLUP_INPUT_DEFAULTS = {
  // Only bundle local package files
  external: (id, parent, isResolved) => {
    // Skip if already handled by plugin
    if (isResolved || parent.includes('?commonjs-proxy')) {
      return false;
    }
    return /^[^./\0]/.test(id);
  },
  plugins: [
    // @ts-ignore
    pluginReplace({
      'process.env.NODE_ENV': `"${process.env.NODE_ENV}"` || '"development"'
    }),
    // @ts-ignore
    pluginNodeResolve({
      mainFields: ['browser', 'module', 'main']
    }),
    // @ts-ignore
    pluginJson(),
    // @ts-ignore
    pluginCommonjs({
      sourceMap: false
    })
  ],
  treeshake: false
};
/** @type { OutputOptions } */
const ROLLUP_OUTPUT_DEFAULTS = {
  format: 'es',
  sourcemap: false
};

/**
 * Bundle package at 'id'
 *
 * @param { string } inputPath
 * @param { string } outputPath
 * @param { RollupOptions } [overrideOptions]
 * @param { (err?: Error) => void } fn
 * @returns { Promise<void> }
 */
var bundleWorker = async function bundle(
  inputPath,
  outputPath,
  overrideOptions,
  fn
) {
  const parsedOverrideOptions = parseOptions(overrideOptions);
  const inputOptions = {
    input: inputPath,
    ...ROLLUP_INPUT_DEFAULTS,
    ...parsedOverrideOptions.input
  };
  const outputOptions = {
    file: outputPath,
    ...ROLLUP_OUTPUT_DEFAULTS,
    ...parsedOverrideOptions.output
  };

  try {
    const bundled = await rollup(inputOptions);
    await bundled.write(outputOptions);
    fn();
  } catch (err) {
    console.log(err);
    fn(err);
  }
};

/**
 * Parse override options
 *
 * @param { RollupOptions } [options]
 * @returns { { input: object, output: object } }
 */
function parseOptions(options) {
  if (!options) {
    return { input: {}, output: {} };
  }

  const { input, treeshake, output = {}, watch, ...inputOverride } = options;
  // @ts-ignore
  const { file, format, sourcemap, ...outputOverride } = output;

  return { input: inputOverride, output: outputOverride };
}

const { find } = file;
const { isRelativeFilePath: isRelativeFilePath$1 } = is;



const MAX_FILE_SYSTEM_DEPTH = 10;
const RE_TRAILING = /\/+$|\\+$/;

var _package = {
  getPackage,
  resolveAliasPath,
  resolvePackagePath
};

/**
 * Retrieve package details for "filePath"
 *
 * @param { string } filePath
 * @param { string } [packagePath]
 * @returns { Package | undefined }
 */
function getPackage(filePath, packagePath = resolvePackagePath(filePath)) {
  if (packagePath === undefined || !fs.existsSync(packagePath)) {
    return;
  }

  const manifestPath = path.join(packagePath, 'package.json');
  const isNodeModule = packagePath !== process.cwd();
  const paths = resolveNodeModules(packagePath);
  /** @type { Package } */
  const pkg = {
    aliases: {},
    isNodeModule,
    manifestPath,
    main: '',
    name: '',
    path: packagePath,
    paths,
    version: ''
  };
  const findOptions = {
    directories: [packagePath, ...pkg.paths],
    type: 'js'
  };

  try {
    const json = JSON.parse(fs.readFileSync(manifestPath, 'utf8'));
    let main = find(json.module || json.main || 'index.js', findOptions);

    // Resolve "browser" aliasing
    if (json.browser) {
      if (typeof json.browser === 'string') {
        main = find(json.browser, findOptions);
      } else {
        for (let key in json.browser) {
          let value = json.browser[key];

          if (typeof value === 'string') {
            // If no extension, or not a relative path, it's a package reference
            if (path.extname(key) || isRelativeFilePath$1(key)) {
              // @ts-ignore
              key = find(key, findOptions);
            }
            if (path.extname(value) || isRelativeFilePath$1(value)) {
              value = find(value, findOptions);
            }

            if (key !== undefined && value !== undefined) {
              if (key === main) {
                main = value;
              }
              pkg.aliases[key] = value;
            } else {
              // TODO: warn unable to resolve
            }
          } else {
            // TODO: warn about disabled module?
          }
        }
      }
    }

    // Store "main" as alias
    if (main) {
      pkg.aliases[packagePath] = main;
      pkg.aliases[json.name] = main;
    }

    pkg.name = json.name;
    pkg.main = main;
    pkg.version = json.version;
  } catch (err) {
    // No package.json found
  }

  return pkg;
}

/**
 * Resolve alias for "filePath"
 *
 * @param { string } filePath
 * @param { Package } pkg
 * @returns { string }
 */
function resolveAliasPath(filePath, pkg) {
  // Follow chain of aliases
  // a => b; b => c; c => d
  while (filePath in pkg.aliases) {
    filePath = pkg.aliases[filePath];
  }

  return filePath;
}

/**
 * Retrieve path to nearest directory with package.json for "filePath"
 *
 * @param { string } filePath
 * @returns { string | undefined }
 */
function resolvePackagePath(filePath) {
  filePath = filePath.replace(RE_TRAILING, '');
  const cwd = process.cwd();
  const isNodeModule = filePath.includes('node_modules');
  let depth = MAX_FILE_SYSTEM_DEPTH;
  let dir = filePath;
  let parent = '';
  let root = cwd;

  // Set root to first node_modules dir if in node_modules
  if (isNodeModule) {
    const parts = filePath.split(path.sep);
    const idx = parts.indexOf('node_modules');

    root = parts.slice(0, idx + 1).join(path.sep);
  }

  while (true) {
    // Stop at directory with package.json
    if (fs.existsSync(path.join(dir, 'package.json'))) {
      return dir;
    }

    parent = path.dirname(dir);

    // Stop if we hit root or max file system depth
    // Convert to lowercase to fix problems on Windows
    if (
      dir === root ||
      !--depth ||
      parent.toLowerCase() === dir.toLowerCase()
    ) {
      break;
    }

    // Walk
    dir = parent;
  }
}

/**
 * Gather all node_modules directories reachable from "pkgPath"
 *
 * @param { string } pkgPath
 * @returns { Array<string> }
 */
function resolveNodeModules(pkgPath) {
  let dir = pkgPath;
  /** @type { Array<string> } */
  let dirs = [];
  let depth = MAX_FILE_SYSTEM_DEPTH;
  let parent;

  if (process.env.NODE_PATH !== undefined) {
    dirs = process.env.NODE_PATH.split(path.delimiter).map((dir) =>
      path.resolve(dir)
    );
  }

  while (true) {
    parent = path.dirname(dir);
    // Stop if we hit max file system depth or root
    // Convert to lowercase to fix problems on Windows
    if (!--depth || parent.toLowerCase() === dir.toLowerCase()) {
      break;
    }

    const nodeModulesPath = path.resolve(dir, 'node_modules');

    if (fs.existsSync(nodeModulesPath)) {
      dirs.push(nodeModulesPath);
    }

    // Walk
    dir = parent;
  }

  return dirs;
}

const {
  getPackage: getPackage$1,
  resolveAliasPath: resolveAliasPath$1,
  resolvePackagePath: resolvePackagePath$1
} = _package;
const {
  find: find$1,
  getProjectPath,
  resolveRealFilePath
} = file;
const { isAbsoluteFilePath: isAbsoluteFilePath$1, isRelativeFilePath: isRelativeFilePath$2 } = is;


const cache = new Map();
const packageCache = new Map();

var resolver = {
  clearResolverCache,
  getCachedPackage,
  resolve
};

/**
 * Resolve file path for "id" relative to "fromFilePath",
 * where "id" can be an absolute path, relative path, or npm package id
 *
 * @param { string } id
 * @param { string } [fromFilePath]
 * @returns { string | undefined }
 */
function resolve(id, fromFilePath = path.resolve('index.js')) {
  if (!id) {
    return;
  }

  const key = getCacheKey(fromFilePath, id);

  if (cache.has(key)) {
    return cache.get(key);
  }

  const filePath = doResolve(id, path.dirname(fromFilePath));

  if (!filePath) {
    return;
  }

  cache.set(key, filePath);
  return filePath;
}

/**
 * Retrieve file path for "id" relative to "fromDirPath"
 *
 * @param { string } id
 * @param { string } fromDirPath
 * @returns { string | undefined }
 */
function doResolve(id, fromDirPath) {
  const pkg = getCachedPackage(fromDirPath);

  if (!pkg) {
    return;
  }

  /** @type { string | undefined } */
  let filePath = resolveAliasPath$1(
    isRelativeFilePath$2(id) ? path.join(fromDirPath, id) : id,
    pkg
  );

  if (isAbsoluteFilePath$1(filePath)) {
    // @ts-ignore
    filePath = resolveAliasPath$1(find$1(filePath, { type: 'js' }), pkg);

    // Allow package references to fall through (isAbsoluteFilePath ==> false)
    if (!filePath || isAbsoluteFilePath$1(filePath)) {
      return resolveRealFilePath(filePath);
    }
  }

  // filePath must be a package reference, so restart search from each package dir working upwards
  id = filePath;

  for (const packageDirPath of pkg.paths) {
    if (fromDirPath !== packageDirPath) {
      filePath = path.join(packageDirPath, id);
      filePath = doResolve(filePath, filePath);

      if (filePath) {
        return resolveRealFilePath(filePath);
      }
    }
  }
}

/**
 * Retrieve cache key
 *
 * @param { string } fromPath
 * @param { string } id
 * @returns { string }
 */
function getCacheKey(fromPath, id) {
  return `${getProjectPath(fromPath)}:${id}`;
}

/**
 * Retrieve Package instance for "dir"
 *
 * @param { string } dir
 * @returns { Package }
 */
function getCachedPackage(dir) {
  const pkgPath = resolvePackagePath$1(dir);
  let pkg = packageCache.get(pkgPath);

  if (!pkg) {
    pkg = getPackage$1(dir, pkgPath);
    packageCache.set(pkgPath, pkg);
  }

  return pkg;
}

/**
 * Clear caches
 */
function clearResolverCache() {
  cache.clear();
  packageCache.clear();
}

/** @type { {[key: string]: { start: [number, number], elapsed: number }} } */
const timers = {};

var stopwatch = {
  start,
  pause,
  stop,
  clear,
  msDiff
};

/**
 * Start timer with 'id'
 *
 * @param { string } id
 */
function start(id) {
  if (!timers[id]) {
    timers[id] = {
      start: [0, 0],
      elapsed: 0
    };
  }
  timers[id].start = process.hrtime();
}

/**
 * Pause timer with 'id'
 *
 * @param { string } id
 */
function pause(id) {
  if (!timers[id]) {
    return start(id);
  }

  timers[id].elapsed += msDiff(process.hrtime(), timers[id].start);
}

/**
 * Stop timer with 'id' and return elapsed
 *
 * @param { string } id
 * @param { boolean } formatted
 * @param { boolean } padded
 * @returns { number | string }
 */
function stop(id, formatted, padded) {
  if (!timers[id]) {
    return formatted ? format(0, padded ? 5 : 0) : 0;
  }

  const elapsed =
    timers[id].elapsed + msDiff(process.hrtime(), timers[id].start);

  clear(id);
  return formatted ? format(elapsed, padded ? 5 : 0) : elapsed;
}

/**
 * clear timer with 'id'
 *
 * @param { string } id
 */
function clear(id) {
  delete timers[id];
}

/**
 * Retrieve difference in ms
 *
 * @param { [number, number] } t1
 * @param { [number, number] } t2
 * @returns { number }
 */
function msDiff(t1, t2) {
  const start = (t1[0] * 1e9 + t1[1]) / 1e6;
  const stop = (t2[0] * 1e9 + t2[1]) / 1e6;
  return Math.ceil((start - stop) * 100) / 100;
}

/**
 * Format 'duration'
 *
 * @param { number } duration
 * @param { number } [length]
 * @returns { string }
 */
function format(duration, length) {
  const colour = duration > 10 ? (duration > 100 ? 'red' : 'yellow') : 'green';
  let formatted =
    duration < 1000
      ? `${Math.round(duration)}ms`
      : `${Math.floor((duration / 1000) * 100) / 100}s`;

  if (length && 'padStart' in String.prototype) {
    formatted = formatted.padStart(length, ' ');
  }

  return chalk[colour](formatted);
}

const { info: info$1, error: error$1 } = log;
const {
  isJsFilePath: isJsFilePath$1,
  isNodeModuleFilePath: isNodeModuleFilePath$1,
  isPromise: isPromise$1
} = is;



const debug = debug$a('dvlp:module');

const { getCachedPackage: getCachedPackage$1 } = resolver;

const { resolve: resolve$1 } = resolver;



/** @type { Map<string, true | Promise<string>> } */
const cache$1 = new Map();
/** @type { import('worker-farm').Workers | undefined } */
let workers;

if (config.maxModuleBundlerWorkers) {
  debug(`bundling modules with ${config.maxModuleBundlerWorkers} workers`);
}

if (fs.existsSync(config.bundleDir)) {
  const names = new Map();

  fs.readdirSync(config.bundleDir)
    .filter(isJsFilePath$1)
    .forEach((resolvedId) => {
      const name = resolvedId.slice(0, resolvedId.lastIndexOf('-'));

      if (!names.has(name)) {
        cache$1.set(resolvedId, true);
        names.set(name, resolvedId);
      } else {
        // Clear instances if duplicates with different versions
        const existing = names.get(name);

        cache$1.delete(existing);
        fs.unlinkSync(path.resolve(config.bundleDir, existing));
        fs.unlinkSync(path.resolve(config.bundleDir, resolvedId));
      }
    });
}

var bundler_1 = {
  bundle,
  cleanBundles,
  destroyWorkers,
  resolveModuleId,
  resolveModulePath
};

/**
 * Resolve module id into cacheable id
 *
 * @param { string } id
 * @param { string } filePath
 * @returns { string }
 */
function resolveModuleId(id, filePath) {
  if (!isNodeModuleFilePath$1(filePath)) {
    return '';
  }

  const pkg = getCachedPackage$1(path.dirname(filePath));

  return `${encodeId(id)}-${pkg.version}.js`;
}

/**
 * Retrieve path to cached 'resolvedId'
 *
 * @param { string } resolvedId
 * @returns { string }
 */
function resolveModulePath(resolvedId) {
  return path.join(config.bundleDir, resolvedId);
}

/**
 * Trigger bundle of 'resolvedId'
 *
 * @param { string } resolvedId
 * @param { string } [originalId]
 * @param { string } [inputPath]
 * @param { import("rollup").RollupOptions } [rollupConfig]
 * @returns { true | Promise<string> | undefined }
 */
function bundle(resolvedId, originalId, inputPath, rollupConfig) {
  if (!resolvedId) {
    return;
  }
  if (!originalId) {
    originalId = decodeId(resolvedId.slice(0, resolvedId.lastIndexOf('-')));
  }
  if (!inputPath) {
    inputPath = resolve$1(originalId);
  }

  const outputPath = resolveModulePath(resolvedId);

  if (!cache$1.has(resolvedId)) {
    return doBundle(
      resolvedId,
      originalId,
      // @ts-ignore
      inputPath,
      outputPath,
      rollupConfig
    );
  } else if (isPromise$1(cache$1.get(resolvedId))) {
    return cache$1.get(resolvedId);
  } else {
    return Promise.resolve(outputPath);
  }
}

/**
 * Bundle module at 'oringinalId'
 *
 * @param { string } resolvedId
 * @param { string } oringinalId
 * @param { string } inputPath
 * @param { string } outputPath
 * @param { import("rollup").RollupOptions } [rollupConfig]
 * @returns { Promise<string> }
 */
function doBundle(
  resolvedId,
  oringinalId,
  inputPath,
  outputPath,
  rollupConfig
) {
  const promiseToCache = new Promise(async (resolve, reject) => {
    stopwatch.start(oringinalId);

    getBundler()(inputPath, outputPath, rollupConfig, (err) => {
      if (err) {
        error$1(`unable to bundle ${oringinalId}`);
        cache$1.delete(resolvedId);
        return reject(err);
      }

      // Can't use file.getProjectPath() here because of circular dependency
      info$1(
        `${stopwatch.stop(oringinalId, true, true)} bundled ${chalk.green(
          oringinalId
        )} as ${chalk.green(path.relative(process.cwd(), outputPath))}`
      );
      cache$1.set(resolvedId, true);
      resolve(outputPath);
    });
  });

  cache$1.set(resolvedId, promiseToCache);
  return promiseToCache;
}

/**
 * Retrieve bundler.
 * Starts workers if config.maxModuleBundlerWorkers > 0,
 * otherwise uses bundler directly
 *
 * @returns { BundleWorker }
 */
function getBundler() {
  if (!config.maxModuleBundlerWorkers) {
    return bundleWorker;
  }

  if (!workers) {
    workers = workerFarm(
      { maxConcurrentWorkers: config.maxModuleBundlerWorkers },
      path.resolve(__dirname, './bundle-worker.js')
    );
    debug(`spawned ${config.maxModuleBundlerWorkers} bundler workers`);
  }

  return workers;
}

/**
 * Clear memory and disk cache
 */
function cleanBundles() {
  for (const resolvedId of cache$1.keys()) {
    try {
      fs.unlinkSync(resolveModulePath(resolvedId));
    } catch (err) {
      // ignore
    } finally {
      cache$1.delete(resolvedId);
    }
  }
}

/**
 * Terminate workers
 */
function destroyWorkers() {
  return new Promise((resolve, reject) => {
    if (!workers) {
      return resolve();
    }

    workerFarm.end(workers, (/** @type { string } */ msg) => {
      workers = undefined;
      if (msg) {
        return reject(Error(msg));
      }
      resolve();
    });
  });
}

/**
 * Encode "id"
 *
 * @param { string } id
 */
function encodeId(id) {
  return id.replace(/\//g, '__');
}

/**
 * Decode "id"
 *
 * @param { string } id
 */
function decodeId(id) {
  return id.replace(/__/g, '/');
}

const debug$1 = debug$a('dvlp:reload');

const { EventSource } = fayeWebsocket;


const { getTypeFromPath } = file;

const { noisyInfo: noisyInfo$1 } = log;


const DEFAULT_CLIENT_CONFIG = {
  headers: { 'Access-Control-Allow-Origin': '*' },
  ping: 15,
  retry: 10
};
const DEFAULT_PORT = 3529;
const ENDPOINT = '/dvlpreload';

const reloadClient =
  '!function(){if("undefined"!=typeof URL&&"undefined"!=typeof EventSource){var e,n,t=new URL(location.origin);t.port=$RELOAD_PORT,t.pathname="/dvlpreload",function o(){(e=new EventSource(t.href)).onopen=function(){n=5},e.onerror=function(){e.close(),n--&&o()},e.addEventListener("reload",(function(){location.reload()})),e.addEventListener("refresh",(function(e){try{for(var n=JSON.parse(e.data).filePath,t=0;t<document.styleSheets.length;t++)if(r(document.styleSheets[t],n))return}catch(e){}location.reload()}))}()}function r(e,n){if(e.href&&o(n,e.href)&&e.ownerNode)return function(e){var n=e.cloneNode(!1),t=e.parentNode;t&&(t.removeChild(e),t.lastChild===e?t.appendChild(n):t.insertBefore(n,e.nextSibling));return!0}(e.ownerNode);for(var t=0;t<e.cssRules.length;t++){var l=e.cssRules[t];if(l.type===CSSRule.IMPORT_RULE)return o(n,l.href,e.href)?i(l,t):r(l.styleSheet,n)}return!1}function o(e,n,t){t=t||location.href;var r=new URL(n,t);return r.origin.includes("localhost")&&r.pathname===e}function i(e,n){var t=e.parentStyleSheet;return t.insertRule(e.cssText,n),t.deleteRule(n+1),!0}}();' ;

/**
 * Create reload server
 *
 * @returns { Promise<Reloader> }
 */
var reloader = async function reloadServer() {
  const port = await getPort({
    port: getPort.makeRange(DEFAULT_PORT, DEFAULT_PORT + 100)
  });
  const server = new ReloadServer(port);
  const client = reloadClient.replace(/\$RELOAD_PORT/g, String(port));

  await server.start();

  return {
    client,
    destroy: server.destroy.bind(server),
    send: server.send.bind(server),
    url: `http://localhost:${server.port}${ENDPOINT}`
  };
};

class ReloadServer {
  /**
   * Constructor
   *
   * @param { number } port
   */
  constructor(port) {
    /** @type { Set<EventSource> } */
    this.clients = new Set();
    this.port = port;
    this.server;
  }

  /**
   * Start server
   *
   * @returns { Promise<void> }
   */
  start() {
    return new Promise((resolve, reject) => {
      /** @type { DestroyableHttpServer } */
      this.server = http.createServer(async (req, res) => {
        // @ts-ignore
        if (req.url !== ENDPOINT || !EventSource.isEventSource(req)) {
          res.writeHead(404);
          res.end();
          return;
        }

        // @ts-ignore
        const client = new EventSource(req, res, DEFAULT_CLIENT_CONFIG);

        this.clients.add(client);
        debug$1('added reload connection', this.clients.size);

        client.on('close', () => {
          this.clients.delete(client);
          debug$1('removed reload connection', this.clients.size);
        });
      });

      serverDestroy(this.server);
      this.server.timeout = this.server.keepAliveTimeout = 0;
      this.server.unref();
      this.server.on('error', reject);
      this.server.on('listening', resolve);

      this.server.listen(this.port);
    });
  }

  /**
   * Send reload message to clients for changed 'filePath'
   *
   * @param { string } filePath
   * @returns { void }
   */
  send(filePath) {
    const type = getTypeFromPath(filePath);
    const event = type === 'css' ? 'refresh' : 'reload';
    const data = JSON.stringify({ type, filePath });

    if (this.clients.size) {
      noisyInfo$1(
        `${chalk.yellow(`  ⟲ ${event}ing`)} ${this.clients.size} client${
          this.clients.size > 1 ? 's' : ''
        }`
      );

      for (const client of this.clients) {
        // @ts-ignore
        client.send(data, { event });
      }
    }
  }

  /**
   * Stop running server
   *
   * @returns { Promise<void> }
   */
  stop() {
    return new Promise((resolve) => {
      if (!this.server) {
        return resolve();
      }

      this.server.removeAllListeners();
      // @ts-ignore
      this.server.destroy(() => {
        debug$1('server stopped');
        resolve();
      });
    });
  }

  /**
   * Destroy instance
   *
   * @returns { Promise<void> }
   */
  destroy() {
    debug$1('destroying');
    for (const client of this.clients) {
      client.close();
    }
    this.clients.clear();
    return this.stop();
  }
}

const { isLocalhost: isLocalhost$1, isProxy: isProxy$1 } = is;
const { URL: URL$1 } = url$1;

/** @type { Set<InterceptClientRequestCallback> } */
const clientRequestListeners = new Set();
/** @type { Set<InterceptFileReadCallback> } */
const fileReadListeners = new Set();
/** @type { Set<InterceptProcessOnCallback> } */
const processOnListeners = new Set();
const originalHttpRequest = http.request;
const originalHttpGet = http.get;
const originalHttpsRequest = https.request;
const originalHttpsGet = https.get;
const originalProcessOn = process.on;
const originalReadStreamRead = fs.ReadStream.prototype._read;
const originalReadFile = fs.readFile;
const originalReadFileSync = fs.readFileSync;

/** @typedef { import("http").ClientRequestArgs & { href?: string } } ClientRequestArgs */

var intercept = {
  interceptClientRequest,
  interceptFileRead,
  interceptProcessOn
};

/**
 * Listen for file system reads and report
 *
 * @param { InterceptFileReadCallback } fn
 * @returns { () => void }
 */
function interceptFileRead(fn) {
  if (!isProxy$1(fs.readFile)) {
    // Proxy ReadStream private method to work around patching by graceful-fs
    const ReadStream = fs.ReadStream.prototype;

    ReadStream._read = new Proxy(ReadStream._read, {
      apply(target, ctx, args) {
        if (notifyListeners(fileReadListeners, ctx.path) === false) {
          return;
        }
        return Reflect.apply(target, ctx, args);
      }
    });

    for (const method of ['readFile', 'readFileSync']) {
      // @ts-ignore
      fs[method] = new Proxy(fs[method], {
        apply(target, ctx, args) {
          if (notifyListeners(fileReadListeners, args[0]) === false) {
            return;
          }
          return Reflect.apply(target, ctx, args);
        }
      });
    }
  }

  fileReadListeners.add(fn);
  return restoreFileRead.bind(null, fn);
}

/**
 * Restore unproxied file reading behaviour
 *
 * @param { InterceptFileReadCallback } fn
 */
function restoreFileRead(fn) {
  fileReadListeners.delete(fn);
  if (!fileReadListeners.size) {
    fs.ReadStream.prototype._read = originalReadStreamRead;
    fs.readFile = originalReadFile;
    fs.readFileSync = originalReadFileSync;
  }
}

/**
 * Listen for client requests
 *
 * @param { InterceptClientRequestCallback } fn
 * @returns { () => void }
 */
function interceptClientRequest(fn) {
  if (!isProxy$1(http.request)) {
    http.request = new Proxy(http.request, {
      apply: clientRequestApplyTrap('http')
    });
    http.get = new Proxy(http.get, { apply: clientRequestApplyTrap('http') });
    https.request = new Proxy(https.request, {
      apply: clientRequestApplyTrap('https')
    });
    https.get = new Proxy(https.get, {
      apply: clientRequestApplyTrap('https')
    });
  }

  clientRequestListeners.add(fn);
  return restoreClientRequest.bind(null, fn);
}

/**
 * Restore unproxied client request behaviour
 *
 * @param { InterceptClientRequestCallback } fn
 */
function restoreClientRequest(fn) {
  clientRequestListeners.delete(fn);
  if (!clientRequestListeners.size) {
    http.request = originalHttpRequest;
    http.get = originalHttpGet;
    https.request = originalHttpsRequest;
    https.get = originalHttpsGet;
  }
}

/**
 * Create client request Proxy apply trap for 'protocol'
 *
 * @param { string } protocol
 * @returns { (target: object, ctx: object, args: [string | ClientRequestArgs]) => Res }
 */
function clientRequestApplyTrap(protocol) {
  return function apply(target, ctx, args) {
    const url = new URL$1(
      typeof args[0] === 'string'
        ? args[0]
        : getHrefFromRequestOptions(args[0], protocol)
    );

    // TODO: pass method/headers
    if (notifyListeners(clientRequestListeners, url) === false) {
      return;
    }

    if (isLocalhost$1(url.hostname)) {
      // Force to http
      url.protocol = 'http:';
      target =
        target === originalHttpsGet || target === originalHttpGet
          ? originalHttpGet
          : originalHttpRequest;
    }
    if (typeof args[0] === 'string') {
      args[0] = url.href;
    } else {
      args[0].protocol = url.protocol;
      args[0].host = url.host;
      args[0].hostname = url.hostname;
      args[0].port = url.port;
      args[0].path = `${url.href.replace(url.origin, '')}`;
      args[0].href = url.href;
      // Force http agent when localhost (due to mocking most likely)
      if (
        args[0].agent &&
        args[0].agent instanceof http.Agent &&
        // @ts-ignore
        args[0].agent.protocol === 'https:' &&
        isLocalhost$1(url.hostname)
      ) {
        // @ts-ignore
        args[0].agent = new http.Agent(args[0].agent.options || {});
      }
    }

    return Reflect.apply(target, ctx, args);
  };
}

/**
 * Retrieve href from 'options'
 *
 * @param { ClientRequestArgs } options
 * @param { string } protocol
 * @returns { string }
 */
function getHrefFromRequestOptions(options, protocol) {
  if (options.href) {
    return options.href;
  }

  let { host = 'localhost', path, port } = options;

  if (!host.includes(':') && port != null) {
    host += `:${port}`;
  }

  return `${protocol}://${host}${path}`;
}

/**
 * Listen for process event registration
 *
 * @param { InterceptProcessOnCallback } fn
 * @returns { () => void }
 */
function interceptProcessOn(fn) {
  if (!isProxy$1(process.on)) {
    process.on = new Proxy(process.on, {
      apply: (target, ctx, args) => {
        if (notifyListeners(processOnListeners, ...args) === false) {
          return;
        }
        return Reflect.apply(target, ctx, args);
      }
    });
  }

  processOnListeners.add(fn);
  return restoreProcessOn.bind(null, fn);
}

/**
 * Restore unproxied process event registration
 *
 * @param { InterceptProcessOnCallback } fn
 */
function restoreProcessOn(fn) {
  processOnListeners.delete(fn);
  if (!processOnListeners.size) {
    process.on = originalProcessOn;
  }
}

/**
 * Notify 'listeners' with 'args'
 *
 * @param { Set<InterceptClientRequestCallback | InterceptFileReadCallback | InterceptProcessOnCallback> } listeners
 * @param { ...unknown } args
 */
function notifyListeners(listeners, ...args) {
  for (const listener of listeners) {
    // @ts-ignore
    if (listener(...args) === false) {
      return false;
    }
  }
}

var scripts = {
  concatScripts,
  getDvlpGlobalString,
  getProcessEnvString,
  hashScript
};

/**
 * Retrieve process.env polyfill
 *
 * @returns { string }
 */
function getProcessEnvString() {
  return `window.process=window.process||{env:{}};window.process.env.NODE_ENV="${process
    .env.NODE_ENV || 'development'}";`;
}

/**
 * Retrieve DVLP global
 *
 * @returns { string }
 */
function getDvlpGlobalString() {
  return 'window.DVLP=true;';
}

/**
 * Concatenate multiple "scripts" into a single string
 *
 * @param { Array<string> } scripts
 * @return { string }
 */
function concatScripts(scripts) {
  return scripts.filter((script) => !!script).join('\n');
}

/**
 * Retrieve sha256 hash of "script"
 *
 * @param { string } script
 * @returns { string }
 */
function hashScript(script) {
  return crypto
    .createHash('sha256')
    .update(script)
    .digest('base64');
}

const { URL: URL$2 } = url$1;

const RE_WEB_SOCKET = /wss?:/;

var url = {
  filePathToUrl,
  getUrl,
  getUrlCacheKey,
  isWebSocketUrl
};

/**
 * Determine if 'url' is a WebSocket
 *
 * @param { URL } url
 * @returns { boolean }
 */
function isWebSocketUrl(url) {
  return RE_WEB_SOCKET.test(url.protocol);
}

/**
 * Retrieve URL instance from 'req'
 *
 * @param { string | { url: string } | URL } req
 * @returns { URL }
 */
function getUrl(req) {
  if (!(req instanceof URL$2)) {
    req = new URL$2(
      // @ts-ignore
      typeof req === 'string' ? decodeURIComponent(req) : req.url,
      `http://localhost:${config.activePort}`
    );
  }
  // Map loopback address to localhost
  if (req.hostname === '127.0.0.1') {
    req.hostname = 'localhost';
  }
  if (req.pathname.endsWith('/')) {
    req.pathname = req.pathname.slice(0, -1);
  }

  return req;
}

/**
 * Retrieve key for 'url'
 *
 * @param { URL } url
 * @returns { string }
 * @private
 */
function getUrlCacheKey(url) {
  // Map loopback address to localhost
  const host = url.host === '127.0.0.1' ? 'localhost' : url.host;
  let key = `${host}${url.pathname}`;

  if (key.endsWith('/')) {
    key = key.slice(0, -1);
  }

  return key;
}

/**
 * Convert file path to valid url
 * Handles platform differences
 *
 * @param { string } filePath
 * @returns { string }
 */
function filePathToUrl(filePath) {
  return encodeURI(filePath.replace(/^[A-Z]:/, '').replace(/\\/g, '/'));
}

const { isWebSocketUrl: isWebSocketUrl$1, getUrl: getUrl$1, getUrlCacheKey: getUrlCacheKey$1 } = url;
const debug$2 = debug$a('dvlp:push');

const { error: error$2 } = log;

const { EventSource: EventSource$1 } = fayeWebsocket;

const DEFAULT_ES_CONFIG = {
  headers: { 'Access-Control-Allow-Origin': '*' },
  ping: 15,
  retry: 10
};
const RE_SOCKETIO_PROTOCOL = /socket\.?io|EIO/;

/**
 * @type { Map<string, Set<PushClient>> }
 */
const cache$2 = new Map();

var pushEvents = {
  connectClient,
  destroyClients,
  pushEvent
};

/**
 * Initialize EventSource/WebSocket client
 *
 * @param { string | PushStream } stream
 * @param { ...any } args
 * @returns { void }
 */
function connectClient(stream, ...args) {
  const { type, url } = getStream(stream);
  const cacheKey = getUrlCacheKey$1(getUrl$1(url));
  const clients = cache$2.get(cacheKey) || new Set();
  /** @type { PushClient } */
  let client;

  if (type === 'ws') {
    const [req, socket, body] = args;
    const isSocketio = RE_SOCKETIO_PROTOCOL.test(req.url);
    const extensionsHeaders = req.headers['Sec-WebSocket-Extensions'];
    const extensions =
      extensionsHeaders && extensionsHeaders.includes('permessage-deflate')
        ? [permessageDeflate]
        : [];

    client = new fayeWebsocket(req, socket, body, [], { extensions });
    client.on('message', (/** @type { { data: string } } */ event) => {
      debug$2('received ws message', event.data);

      // Handle Socket.io channel protocol
      // ex: 40/channel?somequery=foo
      if (isSocketio) {
        const [packetAndChannel] = event.data.split('?');

        // Send separate packet and packet/channel responses
        if (packetAndChannel.includes('/')) {
          client.send(packetAndChannel.slice(0, packetAndChannel.indexOf('/')));
          client.send(packetAndChannel);
        }
      }
    });
    if (isSocketio) {
      client.send(
        '0{"sid":"dvlp","upgrades":[],"pingInterval":250000,"pingTimeout":600000}'
      );
    }
  } else {
    const [req, res] = args;

    // @ts-ignore
    client = new EventSource$1(req, res, DEFAULT_ES_CONFIG);
  }

  clients.add(client);
  cache$2.set(cacheKey, clients);
  debug$2(`added ${type} connection`, clients.size);

  client.on('close', () => {
    clients.delete(client);
    if (!clients.size) {
      cache$2.delete(cacheKey);
    }
    debug$2(`removed ${type} connection`, cache$2.size);
  });
}

/**
 * Push event data to WebSocket/EventSource clients
 *
 * @param { string | PushStream } stream
 * @param { PushEvent } event
 * @returns { void }
 */
function pushEvent(stream, event) {
  if (!stream || !event) {
    return;
  }

  const { url, type } = getStream(stream);
  const cacheKey = getUrlCacheKey$1(getUrl$1(url));
  const clients = cache$2.get(cacheKey);

  if (clients === undefined) {
    return error$2(`no push clients registered for ${url}`);
  }

  let { message, options } = event;

  if (!(typeof message === 'string')) {
    try {
      message = JSON.stringify(message);
    } catch (err) {
      return error$2(`unable to stringify message for push event`, message);
    }
  }
  if (type === 'ws' && options !== undefined) {
    const { event = '', namespace = '/', protocol } = options;

    // Handle socket.io protocol
    // https://github.com/socketio/socket.io-protocol/blob/master/Readme.md
    if (protocol && RE_SOCKETIO_PROTOCOL.test(protocol)) {
      message = `42${namespace},["${event}",${message}]`;
      // message = `${Buffer.from(message).length}:${message}`;
    }
    options = undefined;
  }

  debug$2(
    `push to ${clients.size} client${
      clients.size > 1 ? 's' : ''
    } connected on ${url}`
  );
  debug$2(message);
  for (const client of clients) {
    client.send(message, options);
  }
}

/**
 * Destroy all active push clients for connection at 'url'
 * If 'url' not defined, destroys all clients for all connections
 *
 * @param { string | PushStream } [stream]
 * @returns { void }
 */
function destroyClients(stream) {
  if (stream === undefined) {
    for (const cacheKey of cache$2.keys()) {
      destroyClient(cacheKey);
    }
    return;
  }

  const { url } = getStream(stream);
  const cacheKey = getUrlCacheKey$1(getUrl$1(url));

  destroyClient(cacheKey);
}

/**
 * Destroy client at 'cacheKey'
 *
 * @param { string } cacheKey
 * @returns { void }
 * @private
 */
function destroyClient(cacheKey) {
  const clients = cache$2.get(cacheKey);

  if (clients !== undefined) {
    for (const client of clients) {
      client.removeAllListeners();
      client.close();
    }
    clients.clear();
    cache$2.delete(cacheKey);
  }
}

/**
 * Retrieve PushStream from 'stream'
 * If passed as string, will determine type from url
 *
 * @param { string | PushStream } stream
 * @returns { PushStream }
 * @private
 */
function getStream(stream) {
  if (typeof stream === 'string') {
    const url = getUrl$1(stream);
    const isWebSocket = isWebSocketUrl$1(url);

    stream = {
      url: url.href,
      type: isWebSocket ? 'ws' : 'es'
    };
  }

  return stream;
}

const debug$3 = debug$a('dvlp:fileserver');
const { find: find$2 } = file;

const { info: info$2 } = log;
const { isHtmlRequest: isHtmlRequest$1 } = is;
const { maxAge } = config;



/**
 * Create default file server instance
 */
var fileServer = function createFileServer() {
  // @ts-ignore
  http.createServer(createRequestHandler()).listen(process.env.PORT);
};

/**
 * Create request handler
 *
 * @returns { RequestHandler }
 */
function createRequestHandler() {
  return function requestHandler(req, res) {
    const url = req.url;

    res.once('finish', () => {
      const duration = stopwatch.stop(url, true, true);
      const reroute =
        url !== req.url ? `(re-routed to ${chalk.green(req.url)})` : '';

      // 'transpiled' is added by transpile utility if handled there
      if (!res.transpiled) {
        info$2(
          res.statusCode < 300
            ? `${duration} handled request for ${chalk.green(url)} ${reroute}`
            : `${duration} [${
                res.statusCode
              }] unhandled request for ${chalk.red(url)} ${reroute}`
        );
      }
    });

    let filePath = find$2(req);

    // Re-write to root index.html
    if (!filePath && isHtmlRequest$1(req)) {
      req.url = '/';
      filePath = find$2(req);
    }

    if (!filePath) {
      debug$3(`not found "${req.url}"`);
      res.writeHead(404);
      return res.end();
    }

    debug$3(`sending "${filePath}"`);
    return send(req, filePath, {
      cacheControl: true,
      dotfiles: 'allow',
      maxAge
    }).pipe(res);
  };
}

const { error: error$3, info: info$3, noisyInfo: noisyInfo$2 } = log;
const { isInvalidFilePath: isInvalidFilePath$1, isJsonFilePath: isJsonFilePath$1 } = is;
const { match, pathToRegexp } = pathToRegexp$1;


const debug$4 = debug$a('dvlp:mock');

const { getProjectPath: getProjectPath$1 } = file;
const { getUrl: getUrl$2 } = url;
const { interceptClientRequest: interceptClientRequest$1 } = intercept;





const mockClient =
  "!function(){var e=/wss?:/,n=window.XMLHttpRequest.prototype.open,t=window.fetch,o=$MOCKS.map((function(e){return e.originRegex=new RegExp(e.originRegex),e.pathRegex=new RegExp(e.pathRegex),e})),c=o.reduce((function(e,n){return n.events&&(e[n.href]=n.events),e}),{});window.XMLHttpRequest.prototype.open=function(e,t){return n.call(this,e,h(t))},\"undefined\"!=typeof Proxy&&(\"undefined\"!=typeof fetch&&(window.fetch=new Proxy(window.fetch,{apply:function(e,n,t){return t[0]=h(t[0]),Reflect.apply(e,n,t)}})),\"undefined\"!=typeof EventSource&&(window.EventSource=new Proxy(window.EventSource,{construct:function(e,n){return n[0]=h(n[0]),Reflect.construct(e,n)}})),\"undefined\"!=typeof WebSocket&&(window.WebSocket=new Proxy(window.WebSocket,{construct:function(e,n){return n[0]=h(n[0]),Reflect.construct(e,n)}})));var r=\"color: hotpink; font-size: 14px; font-weight: bold; text-decoration: underline dotted\",i=\"color: hotpink; font-weight: bold; text-decoration: underline dotted\",s=\"color: grey; font-weight: bold\",a=\"color: grey\",p=\"color: grey; font-style: italic\",l=\"color: grey; font-style: italic; font-weight: bold\";function h(n){var t=function(e){try{return new URL(e)}catch(t){var n=document.createElement(\"a\");return n.href=e,{protocol:n.protocol,host:n.host,origin:n.protocol+\"//\"+n.host,pathname:(\"/\"!==n.pathname.charAt(0)?\"/\":\"\")+n.pathname,search:n.search}}}(\"string\"==typeof n?n:n.url);return function(e){for(var n=-1===e.origin.indexOf(e.host)?e.origin+e.host:e.origin,t=0;t<o.length;t++){var c=o[t];if(c.originRegex.test(n)&&(c.search||d(e.search,c.search))&&null!=c.pathRegex.exec(e.pathname))return c}}(t)&&(n=(e.test(t.protocol)?\"ws:\":location.protocol)+location.host+location.pathname+\"?dvlpmock=\"+encodeURIComponent(n)+(location.search?\"&\"+location.search.slice(1):\"\")),n}function d(e,n){var t=u(e),o=u(n);if(Object.keys(t).length!==Object.keys(o).length)return!1;for(var c in t){var r=t[c],i=o[c];if(r.length!==i.length)return!1;for(var s=0;s<r.length;s++){var a=r[s];if(-1===i.indexOf(a))return!1}}return!0}function u(e){e=e.slice(1).split(\"&\");for(var n={},t=0;t<e.length;t++){var o=e[t].split(\"=\"),c=o[0],r=o[1];c in n||(n[c]=[]),n[c].push(r)}return n}window.dvlp={events:c,pushEvent:function(e,n){t(\"/dvlp/push-event\",{method:\"POST\",headers:{Accept:\"application/json\",\"Content-Type\":\"application/json\"},body:JSON.stringify({stream:e,event:n})})}},Object.defineProperty(window.dvlp,\"help\",{get:function(){console.log(\"\\n%cWelcome to the dvlp client command-line interface!\\n\\n%cUsage:%c dvlp.<command>\\n\\n%cCommands:\\n%c  pushEvent%c - trigger EventSource/WebSocket event when mocking (%cdvlp.pushEvent.help%c for more)\\n\",r,i,l,i,s,a,p,a)}}),Object.defineProperty(window.dvlp.pushEvent,\"help\",{get:function(){console.log('\\n%cPushEvent: trigger EventSource/WebSocket event when mocking\\n\\n%cUsage:%c dvlp.pushEvent(stream:%c string%c, event:%c string|object%c)\\n\\n%cArguments:\\n%c  stream:%c string%c - the mocked EventSource/WebSocket connection endpoint\\n%c  event:%c string%c - the \"name\" of a mocked event\\n%c  event:%c {message, options}%c - event object\\n%c    message:%c string|object%c - the event payload as string or JSON object\\n%c    [options]:%c {event: string, id: string}%c - optional EventSource \"event\" name and \"id\"\\n\\n%cExamples:\\n%c  // Push mock event named \"newsfeed with 9 items\"\\n  dvlp.pushEvent(\"http://localhost:8000/feed\", \"newsfeed with 9 items\");\\n\\n  // Push event with string message payload\\n  dvlp.pushEvent(\"http://localhost:8000/feed\", {message: \"some news\"});\\n\\n  // Push event with object message payload and EventSource event name\\n  dvlp.pushEvent(\"http://localhost:8000/feed\", {\\n    message: {item: \"some news\"},\\n    options: {event: \"news\"}\\n  });\\n\\n%cSee:%c https://github.com/popeindustries/dvlp#mocking\\n\\n%cCurrently mocked events:\\n%c'+Object.keys(c).map((function(e){return\"  \"+e+\"\\n    \"+c[e].join(\"\\n    \")})).join(\"\\n\")+\"\\n\",r,i,l,p,l,p,l,i,l,p,a,l,p,a,l,p,a,l,p,a,l,p,a,i,p,i,s,i,l)}})}();" ;

var mock_1 = class Mock {
  /**
   * Constructor
   *
   * @implements { MockInstance }
   * @param { string | Array<string> } [filePaths]
   */
  constructor(filePaths) {
    /** @type { Set<MockResponseData | MockStreamData> } */
    this.cache = new Set();
    this.clean = this.clean.bind(this);
    this.client = mockClient;
    this._uninterceptClientRequest;

    if (filePaths) {
      this.load(filePaths);
    }
  }

  /**
   * Add new mock for 'res'
   *
   * @param { string | MockRequest } req
   * @param { MockResponse | MockResponseHandler } res
   * @param { boolean } [once]
   * @param { () => void } [onMockCallback]
   * @returns { void }
   */
  addResponse(req, res, once = false, onMockCallback) {
    const ignoreSearch = (isMockRequest(req) && req.ignoreSearch) || false;
    const filePath =
      (isMockRequest(req) && req.filePath) || path.join(process.cwd(), 'mock');
    const [
      url,
      originRegex,
      pathRegex,
      paramsMatch,
      searchParams
    ] = getUrlSegmentsForMatching(req);
    /** @type { MockResponseDataType } */
    let type = 'json';

    if (typeof res !== 'function') {
      if (!res.body) {
        res = { body: res, headers: {} };
      }
      if (typeof res.body === 'string') {
        type = isInvalidFilePath$1(res.body) ? 'html' : 'file';
      }
    }

    const mock = {
      url,
      originRegex,
      pathRegex,
      paramsMatch,
      searchParams,
      ignoreSearch,
      once,
      type,
      filePath,
      callback: onMockCallback,
      response: res
    };

    if (!this._uninterceptClientRequest) {
      this.enableRequestIntercept();
    }
    this.cache.add(mock);
    debug$4(`adding mocked "${typeof req === 'string' ? req : req.url}"`);
  }

  /**
   * Add new push mock for 'events'
   *
   * @param { string | MockPushStream } stream
   * @param { MockPushEvent | Array<MockPushEvent> } events
   */
  addPushEvents(stream, events) {
    if (!Array.isArray(events)) {
      events = [events];
    }
    const ignoreSearch =
      (isMockPushStream(stream) && stream.ignoreSearch) || false;
    const filePath =
      (isMockPushStream(stream) && stream.filePath) ||
      path.join(process.cwd(), 'mock');
    const [
      url,
      originRegex,
      pathRegex,
      paramsMatch,
      searchParams
    ] = getUrlSegmentsForMatching(stream);
    /** @type { MockStreamDataType } */
    const type = originRegex.source.includes('ws') ? 'ws' : 'es';
    // Default to socket.io protocol for ws
    const protocol = (isMockPushStream(stream) && stream.protocol) || type;
    /** @type { MockStreamData["events"] } */
    const eventsData = {};

    for (const event of events) {
      // Ignore events without a name
      if (event.name) {
        /** @type { MockStreamEventData["options"] } */
        const options = { delay: 0, ...event.options, protocol };

        /** @type { Array<MockStreamEventData> } */
        const sequence = [];

        if (event.sequence) {
          for (const sequenceEvent of event.sequence) {
            sequence.push({
              message: sequenceEvent.message || '',
              options: { delay: 0, ...sequenceEvent.options, protocol }
            });
          }
        } else {
          sequence.push({
            name: event.name,
            message: event.message || '',
            options
          });
        }

        if (options.connect) {
          const connectEvent = eventsData.connect || [];

          connectEvent.push(...sequence);
          eventsData.connect = connectEvent;
        }

        eventsData[event.name] = sequence;
      }
    }

    /** @type { MockStreamData } */
    const mock = {
      url,
      originRegex,
      pathRegex,
      paramsMatch,
      searchParams,
      ignoreSearch,
      filePath,
      type,
      protocol,
      events: eventsData
    };

    this.cache.add(mock);
    debug$4(
      `adding mocked stream "${
        typeof stream === 'string' ? stream : stream.url
      }"`
    );
  }

  /**
   * Load mock files from disk
   *
   * @param { string | Array<string> } filePaths
   */
  load(filePaths) {
    if (!Array.isArray(filePaths)) {
      filePaths = [filePaths];
    }

    for (let filePath of filePaths) {
      filePath = path.resolve(filePath);

      try {
        const stat = fs.statSync(filePath);

        if (stat.isDirectory()) {
          this.loadDirectory(filePath);
        } else {
          this.loadFile(filePath);
        }
      } catch (err) {
        error$3(`unable to find mock file ${filePath}`);
      }
    }

    // Client mocking only relevant for loaded mocks
    this.client = mockClient.replace(
      /\$MOCKS/g,
      JSON.stringify(
        Array.from(this.cache).map((mockData) => {
          const data = {
            href: mockData.url.href,
            originRegex: mockData.originRegex.source,
            pathRegex: mockData.pathRegex.source,
            search: mockData.searchParams.toString(),
            ignoreSearch: mockData.ignoreSearch
          };

          if (isMockStreamData(mockData)) {
            // @ts-ignore
            data.events = Object.keys(mockData.events);
          }

          return data;
        }),
        undefined,
        2
      )
    );
  }

  /**
   * Match and handle mock response for 'key'
   * Will respond if 'res' passed
   *
   * @param { string } href
   * @param { Req } [req]
   * @param { Res } [res]
   * @returns { boolean | MockResponseData | undefined }
   */
  matchResponse(href, req, res) {
    const mock = this.getMockData(href);

    if (!mock || !isMockResponseData(mock)) {
      return false;
    }

    if (!req || !res) {
      return mock;
    }
    if (mock.once) {
      this.remove(mock);
    }
    if (mock.callback) {
      mock.callback();
    }
    if (typeof mock.response === 'function') {
      const url = getUrl$2(href);
      const matchObj = mock.paramsMatch(url.pathname);

      req.params = matchObj ? matchObj.params : {};

      return mock.response(req, res);
    }

    const {
      response: { hang, error, missing, offline }
    } = mock;

    // Handle special status
    if (hang) {
      return;
    } else if (error || missing) {
      const statusCode = error ? 500 : 404;
      const body = error ? 'error' : 'missing';

      res.writeHead(statusCode);
      res.end(body);
      return;
    } else if (offline && req) {
      req.socket.destroy();
      return;
    }

    debug$4(`sending mocked "${href}"`);

    const {
      filePath,
      response: { body, headers = {} },
      type
    } = mock;
    let content = body;

    switch (type) {
      case 'file':
        res.setHeader('Access-Control-Allow-Origin', '*');
        // Body is path to file (relative to mock file)
        send(
          {
            // @ts-ignore
            url: href,
            headers: {}
          },
          // @ts-ignore
          path.resolve(path.dirname(filePath), body),
          {
            cacheControl: true,
            dotfiles: 'allow',
            maxAge: config.maxAge
          }
        ).pipe(res);
        return;
      case 'json':
        content = JSON.stringify(body);
        break;
    }

    // @ts-ignore
    res.writeHead(200, {
      // Allow type to be overwritten
      'Content-Type': mime.getType(type),
      ...headers,
      // @ts-ignore
      'Content-Length': Buffer.byteLength(content),
      'Access-Control-Allow-Origin': '*',
      // Overwrite Date
      Date: new Date().toUTCString()
    });
    res.end(content);

    noisyInfo$2(
      `${stopwatch.stop(
        res.url,
        true,
        true
      )} handled mocked request for ${chalk.green(decodeURIComponent(href))}`
    );

    return true;
  }

  /**
   * Match and handle mock push event for 'stream'
   *
   * @param { string | MockPushStream } stream
   * @param { string } name
   * @param { (stream: string | PushStream, event: PushEvent) => void } push
   * @returns { boolean }
   */
  matchPushEvent(stream, name, push) {
    const mock = this.getMockData(stream);

    if (!mock || !isMockStreamData(mock)) {
      return false;
    }

    const eventSequence = mock.events[name];

    if (!eventSequence) {
      return false;
    }

    stopwatch.start(name);
    triggerEventSequence(stream, eventSequence, push).then(() => {
      noisyInfo$2(
        `${stopwatch.stop(
          name,
          true,
          true
        )} triggered mocked push event ${chalk.green(name)}`
      );
    });

    return true;
  }

  /**
   * Determine if 'reqOrMockData' matches cached mock data
   *
   * @param { string | URL | { url: string } | MockResponseData | MockStreamData } reqOrMockData
   * @returns { boolean }
   */
  hasMatch(reqOrMockData) {
    if (isMockResponseData(reqOrMockData) || isMockStreamData(reqOrMockData)) {
      return this.cache.has(reqOrMockData);
    }
    return this.getMockData(reqOrMockData) !== undefined;
  }

  /**
   * Remove existing mock data
   *
   * @param { string | URL | { url: string } | MockResponseData | MockStreamData } reqOrMockData
   * @returns { void }
   */
  remove(reqOrMockData) {
    if (isMockResponseData(reqOrMockData) || isMockStreamData(reqOrMockData)) {
      this.cache.delete(reqOrMockData);
    } else {
      const mockData = this.getMockData(reqOrMockData);

      if (mockData) {
        this.cache.delete(mockData);
      }
    }

    if (this._uninterceptClientRequest && !this.cache.size) {
      this._uninterceptClientRequest();
      this._uninterceptClientRequest = undefined;
    }
  }

  /**
   * Clear all mock data
   *
   * @returns { void }
   */
  clean() {
    this.cache.clear();
    this._uninterceptClientRequest && this._uninterceptClientRequest();
  }

  /**
   * Initialize request interception
   *
   * @private
   */
  enableRequestIntercept() {
    this._uninterceptClientRequest = interceptClientRequest$1((url) => {
      if (this.hasMatch(url.href)) {
        url.searchParams.append('dvlpmock', encodeURIComponent(url.href));
        // Reroute to active server
        url.host = `localhost:${config.activePort}`;
      }
      return true;
    });
  }

  /**
   * Retrieve mock response/stream data
   *
   * @param { string | URL | { url: string } } req
   * @returns { MockResponseData | MockStreamData | undefined }
   * @private
   */
  getMockData(req) {
    const url = getUrl$2(req);

    for (const mock of this.cache) {
      if (
        !mock.originRegex.test(url.origin) ||
        (!mock.ignoreSearch &&
          !isEqualSearchParams(url.searchParams, mock.searchParams))
      ) {
        continue;
      }

      if (mock.pathRegex.exec(url.pathname) != null) {
        return mock;
      }
    }
  }

  /**
   * Load directory at 'dirpath'
   *
   * @param { string } dirpath
   * @private
   */
  loadDirectory(dirpath) {
    fs.readdirSync(dirpath).forEach((filePath) => {
      this.load(path.join(dirpath, filePath));
    });
  }

  /**
   * Load file at 'filePath'
   *
   * @param { string } filePath
   * @private
   */
  loadFile(filePath) {
    if (!isJsonFilePath$1(filePath)) {
      return;
    }

    try {
      /** @type { Array<MockResponseJSONSchema | MockPushEventJSONSchema> } */
      let mocks = JSON.parse(fs.readFileSync(filePath, 'utf8'));
      let count = 0;
      let type = '';

      if (!Array.isArray(mocks)) {
        mocks = [mocks];
      }

      for (const mock of mocks) {
        if (!isMockJSONSchema(mock)) {
          return;
        } else if (!isValidJSONSchema(mock)) {
          return error$3(`invalid mock format for ${filePath}`);
        }

        // @ts-ignore
        const { request, response, stream, events } = mock;

        if (request && response) {
          request.filePath = filePath;
          this.addResponse(request, response, false);
          type = type === 'event' ? 'event/response' : 'response';
          count++;
        } else if (stream && events) {
          stream.filePath = filePath;
          this.addPushEvents(stream, events);
          type = type === 'response' ? 'event/response' : 'event';
          count += events.length;
        }
      }

      info$3(
        `${chalk.green('✔')} loaded ${count} mock ${type}${
          count > 1 ? 's' : ''
        } from ${chalk.green(getProjectPath$1(filePath))}`
      );
    } catch (err) {
      error$3(err);
    }
  }
};

/**
 * Retrieve origin, path regex, and search params for "req"
 *
 * @param { string | MockRequest } req
 * @param { boolean } ignoreSearch
 * @returns { [URL, RegExp, RegExp, import('path-to-regexp').MatchFunction, URLSearchParams] }
 */
function getUrlSegmentsForMatching(req, ignoreSearch) {
  let href = typeof req === 'string' ? req : req.url;

  if (href.includes('127.0.0.1')) {
    href = href.replace('127.0.0.1', 'localhost');
  }

  const url = new URL(href, `http://localhost:${config.activePort}`);
  // Allow matching of both secure/unsecure protocols
  const origin = new RegExp(
    url.origin
      .replace(/http:|https:/, 'https?:')
      .replace('ws:', 'wss?:')
      .replace('//', '\\/\\/')
  );
  let pathname = href.replace(origin, '');
  let search = '';

  if (pathname.includes('?')) {
    const isDynamic = /[:(]/.test(pathname);
    // Queries are escaped with '\\'
    // Strip all preceeding '\', no matter how many, because escaping is hard
    const regex = isDynamic ? /\\+\?(.*)$/ : /\?(.*)$/;
    const match = regex.exec(pathname);

    if (match) {
      search = match[1];
      pathname = pathname.replace(match[0], '');
    }
  }

  return [
    url,
    origin,
    pathToRegexp(pathname),
    match(pathname, { decode: decodeURIComponent }),
    new URLSearchParams(search)
  ];
}

/**
 * Determine if search params are equal
 *
 * @param { URLSearchParams } params1
 * @param { URLSearchParams } params2
 * @returns { boolean }
 */
function isEqualSearchParams(params1, params2) {
  // @ts-ignore
  const keys1 = Array.from(params1.keys());
  // @ts-ignore
  const keys2 = Array.from(params2.keys());

  if (keys1.length !== keys2.length) {
    return false;
  }

  for (const key of keys1) {
    const values1 = params1.getAll(key);
    const values2 = params2.getAll(key);

    if (values1.length !== values2.length) {
      return false;
    }

    for (const value of values1) {
      if (!values2.includes(value)) {
        return false;
      }
    }
  }

  return true;
}

/**
 * Match and handle mock push event for 'stream'
 *
 * @param { string | MockPushStream } stream
 * @param { Array<MockStreamEventData> } eventSequence
 * @param { (stream: string | PushStream, event: PushEvent) => void } push
 * @returns { Promise<void> }
 * @private
 */
async function triggerEventSequence(stream, eventSequence, push) {
  for (const event of eventSequence) {
    const {
      message,
      options: { delay = 0, ...options }
    } = event;

    await sleep(delay);
    push(stream, { message, options });
  }
}

/**
 * Sleep for 'duration' milliseconds
 *
 * @param { number } duration
 * @returns { Promise<void> }
 * @private
 */
function sleep(duration) {
  return new Promise((resolve) => setTimeout(resolve, duration));
}

/**
 *
 * @param { object } json
 * @returns { boolean }
 */
function isMockJSONSchema(json) {
  return (
    ('request' in json && 'response' in json) ||
    ('stream' in json && 'events' in json)
  );
}

/**
 * Validate that 'json' is correct format
 *
 * @param { object } json
 * @returns { boolean }
 */
function isValidJSONSchema(json) {
  return (
    ('request' in json &&
      'url' in json.request &&
      'response' in json &&
      'body' in json.response) ||
    ('stream' in json && 'url' in json.stream && 'events' in json)
  );
}

/**
 * Determine if "req" is a MockRequest
 *
 * @param { object } req
 * @returns { req is MockRequest }
 */
function isMockRequest(req) {
  return (
    req &&
    typeof req === 'object' &&
    req.url !== undefined &&
    req.type === undefined
  );
}

/**
 * Determine if "req" is a MockPushStream
 *
 * @param { object } req
 * @returns { req is MockPushStream }
 */
function isMockPushStream(req) {
  return (
    req &&
    typeof req === 'object' &&
    req.url !== undefined &&
    req.type !== undefined
  );
}

/**
 * Determine if "mock" is a MockResponseData
 *
 * @param { object } mock
 * @returns { mock is MockResponseData }
 */
function isMockResponseData(mock) {
  return mock && typeof mock === 'object' && 'once' in mock;
}

/**
 * Determine if "mock" is a MockStreamData
 *
 * @param { object } mock
 * @returns { mock is MockStreamData }
 */
function isMockStreamData(mock) {
  return mock && typeof mock === 'object' && 'events' in mock;
}

const { bundle: bundle$1, resolveModuleId: resolveModuleId$1 } = bundler_1;
const { getAbsoluteProjectPath, getProjectPath: getProjectPath$2 } = file;
const {
  isCssRequest: isCssRequest$1,
  isHtmlRequest: isHtmlRequest$2,
  isJsRequest: isJsRequest$1,
  isNodeModuleFilePath: isNodeModuleFilePath$2,
  isModule: isModule$1,
  isModuleBundlerFilePath: isModuleBundlerFilePath$1,
  isRelativeFilePath: isRelativeFilePath$3
} = is;
const { warn: warn$1, WARN_BARE_IMPORT: WARN_BARE_IMPORT$1 } = log;

const debug$5 = debug$a('dvlp:patch');
const { filePathToUrl: filePathToUrl$1 } = url;

const { resolve: resolve$2 } = resolver;
const { unzipSync } = zlib;

const RE_CLOSE_BODY_TAG = /<\/body>/i;
const RE_IMPORT = /((?:\bimport\b[^'"]+|\bexport\b[^'"]+\sfrom\s)['"])([^'"\n]+)(['"])/gm;
const RE_NONCE_SHA = /nonce-|sha\d{3}-/;
const RE_OPEN_HEAD_TAG = /<head>/i;

var patch = {
  patchResponse
};

/**
 * Patch response body.
 * Injects reload script into html response,
 * and resolves import ids for js/css response.
 *
 * @param { string } filePath
 * @param { Req } req
 * @param { Res } res
 * @param { PatchResponseConfig } [options]
 */
function patchResponse(
  filePath,
  req,
  res,
  // @ts-ignore
  { rollupConfig, footerScript, headerScript } = {}
) {
  // req.filepath set after file.find(), filepath passed if cached
  filePath = req.filePath || filePath || req.url;
  debug$5(`patching response for "${getProjectPath$2(filePath)}"`);
  proxySetHeader(
    res,
    disableContentEncodingHeader.bind(disableContentEncodingHeader, res)
  );
  if (isHtmlRequest$2(req)) {
    const urls = [];
    const hashes = [];
    const scripts = {
      header: '',
      footer: ''
    };

    if (footerScript) {
      if (footerScript.url) {
        urls.push(footerScript.url);
      }
      if (footerScript.hash) {
        hashes.push(footerScript.hash);
      }
      scripts.footer = footerScript.string;
    }
    if (headerScript) {
      if (headerScript.url) {
        urls.push(headerScript.url);
      }
      if (headerScript.hash) {
        hashes.push(headerScript.hash);
      }
      scripts.header = headerScript.string;
    }
    proxySetHeader(res, injectCSPHeader.bind(injectCSPHeader, urls, hashes));
    proxyBodyWrite(res, (html) => {
      enableCrossOriginHeader(res);
      disableCacheControlHeader(res, req.url);
      return injectScripts(scripts, html);
    });
  } else if (isCssRequest$1(req)) {
    // Disable gzip
    proxyBodyWrite(res, (css) => {
      enableCrossOriginHeader(res);
      disableCacheControlHeader(res, req.url);
      return css;
    });
  } else if (isJsRequest$1(req)) {
    proxyBodyWrite(res, (code) => {
      enableCrossOriginHeader(res);
      disableCacheControlHeader(res, req.url);
      return rewriteImports(filePath, rollupConfig, code);
    });
  }
}

/**
 * Disable Content-Encoding headers
 *
 * @param { Res } res
 * @param { string } headerKey
 * @param { string } headerValue
 * @returns { string | undefined }
 */
function disableContentEncodingHeader(res, headerKey, headerValue) {
  const key = headerKey.toLowerCase();

  if (key === 'content-encoding') {
    res.encoding = headerValue;
    return;
  }

  return headerValue;
}

/**
 * Disable Cache-Control, Content-Encoding headers
 *
 * @param { Res } res
 * @param { string } url
 * @returns { void }
 */
function disableCacheControlHeader(res, url) {
  if (!res.headersSent) {
    if (!isNodeModuleFilePath$2(url) && !isModuleBundlerFilePath$1(url)) {
      res.setHeader('cache-control', 'no-cache, dvlp-disabled');
    }
  }
}

/**
 * Enable Access-Control-Allow-Origin header
 *
 * @param { Res } res
 */
function enableCrossOriginHeader(res) {
  if (!res.headersSent) {
    res.setHeader('Access-Control-Allow-Origin', '*');
  }
}

/**
 * Inject header/footer script tags into 'data'
 *
 * @param { { footer: string, header: string } } scripts
 * @param { string } html
 * @returns { string }
 */
function injectScripts(scripts, html) {
  const { footer, header } = scripts;

  if (header && RE_OPEN_HEAD_TAG.test(html)) {
    debug$5('injecting header script');
    html = html.replace(RE_OPEN_HEAD_TAG, `<head>\n<script>${header}</script>`);
  }
  if (footer && RE_CLOSE_BODY_TAG.test(html)) {
    debug$5('injecting footer script');
    html = html.replace(
      RE_CLOSE_BODY_TAG,
      `<script>${footer}</script>\n</body>`
    );
  }

  return html;
}

/**
 * Inject CSP headers allowing inline script tag
 *
 * @param { Array<string> } urls
 * @param { Array<string> } hashes
 * @param { string } key
 * @param { string } value
 * @returns { string }
 */
function injectCSPHeader(urls, hashes, key, value) {
  const lcKey = key.toLowerCase();

  if (
    lcKey === 'content-security-policy-report-only' ||
    lcKey === 'content-security-policy'
  ) {
    const urlsString = urls.join(' ');
    const hashesString = hashes.map((hash) => `'sha256-${hash}'`).join(' ');
    const rules = value
      .split(';')
      .map((ruleString) => ruleString.trim())
      .reduce((/** @type { {[key: string]: string} } */ rules, ruleString) => {
        const firstSpace = ruleString.indexOf(' ');

        rules[ruleString.slice(0, firstSpace)] = ruleString.slice(
          firstSpace + 1
        );
        return rules;
      }, {});

    if (rules['connect-src']) {
      rules['connect-src'] = `${rules['connect-src']} ${urlsString}`;
    } else {
      rules['connect-src'] = urlsString;
    }
    if (
      rules['script-src'] &&
      (RE_NONCE_SHA.test(value) || !value.includes('unsafe-inline'))
    ) {
      rules['script-src'] = `${rules['script-src']} ${hashesString}`;
    }
    if (rules['default-src'] === "'none'") {
      rules['default-src'] = `* ${hashesString}`;
    }

    value = Object.keys(rules).reduce((value, name) => {
      value += `${name} ${rules[name]}; `;
      return value;
    }, '');
  }

  return value;
}

/**
 * Rewrite bare import references in 'data'
 *
 * @param { string } filePath
 * @param { object } rollupConfig
 * @param { string } code
 * @returns { string }
 */
function rewriteImports(filePath, rollupConfig, code) {
  const projectFilePath = getProjectPath$2(filePath);
  /** @type { {[key: string]: string} } */
  const rewritten = {};
  let match;

  if (!RE_IMPORT.test(code)) {
    debug$5(`no imports to rewrite in "${projectFilePath}"`);
    return code;
  }

  RE_IMPORT.lastIndex = 0;
  while ((match = RE_IMPORT.exec(code))) {
    const [context, pre, id, post] = match;
    const importPath = resolve$2(id, getAbsoluteProjectPath(filePath));
    if (importPath) {
      let newId = '';

      // Bundle if in node_modules and not an es module
      if (isNodeModuleFilePath$2(importPath) && !isModule$1(importPath)) {
        const resolvedId = resolveModuleId$1(id, importPath);

        // Trigger bundling in background while waiting for eventual request
        bundle$1(resolvedId, id, importPath, rollupConfig);
        newId = `/${path.join(config.bundleDirName, resolvedId)}`;
        warn$1(WARN_BARE_IMPORT$1, id);
      } else {
        // Don't rewrite if no change after resolving
        newId =
          isRelativeFilePath$3(id) &&
          path.join(path.dirname(filePath), id) === importPath
            ? id
            : importPath;
      }

      newId = filePathToUrl$1(newId);
      debug$5(`rewrote import id from "${id}" to "${newId}"`);
      rewritten[context] = `${pre}${newId}${post}`;
    } else {
      warn$1(`⚠️  unable to resolve path for "${id}" from "${projectFilePath}"`);
    }
  }

  for (const importString in rewritten) {
    code = code.replace(importString, rewritten[importString]);
  }

  return code;
}

/**
 * Proxy set header for 'res', performing 'action' on write()/end()
 *
 * @param { Res } res
 * @param { (key: string, value: string) => string | undefined } action
 */
function proxySetHeader(res, action) {
  res.setHeader = new Proxy(res.setHeader, {
    apply(target, ctx, args) {
      let [key, value] = args;

      value = action(key, value);

      if (value) {
        return Reflect.apply(target, ctx, [key, value]);
      }
    }
  });

  res.writeHead = new Proxy(res.writeHead, {
    apply(target, ctx, args) {
      // First argument is always statusCode
      if (args.length > 1) {
        const headers = args[args.length - 1];

        for (const key in headers) {
          const value = action(key, headers[key]);

          if (value) {
            headers[key] = value;
          }
        }
      }

      return Reflect.apply(target, ctx, args);
    }
  });
}

/**
 * Proxy body write for 'res', performing 'action' on write()/end()
 *
 * @param { Res } res
 * @param { (data: string) => string } action
 */
function proxyBodyWrite(res, action) {
  const originalSetHeader = res.setHeader;
  /** @type { Buffer } */
  let buffer;

  // Proxy write() to buffer streaming response
  res.write = new Proxy(res.write, {
    apply(target, ctx, args) {
      let [chunk] = args;

      if (typeof chunk === 'string') {
        chunk = Buffer.from(chunk);
      }
      buffer = Buffer.concat([buffer || Buffer.from(''), chunk]);
      return;
    }
  });

  // Proxy end() to intercept response body
  res.end = new Proxy(res.end, {
    apply(target, ctx, args) {
      let [data = buffer, ...extraArgs] = args;
      let size = 0;

      if (data) {
        if (Buffer.isBuffer(data)) {
          if (res.encoding === 'gzip') {
            data = unzipSync(data);
          }
          data = data.toString();
        }
        data = action(data);
        size = Buffer.byteLength(data);
      }

      if (!res.headersSent) {
        if (size) {
          debug$5(`setting Content-Length to ${size}`);
          originalSetHeader.call(res, 'Content-Length', size);
        }
      }

      return Reflect.apply(target, ctx, [data, ...extraArgs]);
    }
  });

  // Prevent setting of Content-Length
  res.setHeader = new Proxy(res.setHeader, {
    apply(target, ctx, args) {
      let [key, value] = args;

      if (key.toLowerCase() === 'content-length') {
        debug$5(`prevented setting Content-Length to ${value}`);
        return;
      }

      return Reflect.apply(target, ctx, [key, value]);
    }
  });

  // Prevent setting of Content-Length
  res.writeHead = new Proxy(res.writeHead, {
    apply(target, ctx, args) {
      // First argument is always statusCode
      if (args.length > 1) {
        for (const key in args[args.length - 1]) {
          if (key.toLowerCase() === 'content-length') {
            debug$5(
              `prevented setting Content-Length to ${
                args[args.length - 1][key]
              }`
            );
            delete args[args.length - 1][key];
          }
        }
      }

      return Reflect.apply(target, ctx, args);
    }
  });
}

const { getProjectPath: getProjectPath$3, getTypeFromPath: getTypeFromPath$1 } = file;

const debug$6 = debug$a('dvlp:transpile');
const { error: error$4, info: info$4 } = log;



/**
 * Transpile content for filePath
 *
 * @param { string } filePath
 * @param { Res } res
 * @param { TranspilerState } state
 * @returns { Promise<void> }
 */
var transpile = async function transpile(filePath, res, state) {
  const { transpilerCache, lastChanged, transpiler } = state;
  const relativeFilePath = getProjectPath$3(filePath);
  // Dependencies that are concatenated during transpile aren't cached,
  // but they are watched when read from file system during transpilation,
  // so transpile again if same type
  const lastChangedIsDependency =
    !!lastChanged &&
    !transpilerCache.has(lastChanged) &&
    getTypeFromPath$1(lastChanged) === getTypeFromPath$1(filePath);
  let content = transpilerCache.get(filePath);
  let transpiled = false;

  if (lastChangedIsDependency || lastChanged === filePath || !content) {
    try {
      content = await transpiler(filePath, false);
      if (content !== undefined) {
        transpiled = true;
        transpilerCache.set(filePath, content);
      }
    } catch (err) {
      debug$6(`error transpiling "${relativeFilePath}"`);
      res.writeHead(500);
      res.end(err.message);
      error$4(err);
      return;
    }
  }

  if (content !== undefined) {
    debug$6(
      `${
        transpiled ? 'transpiled content for' : 'skipping transpile for'
      } "${relativeFilePath}"`
    );
    res.transpiled = true;
    // @ts-ignore
    res.writeHead(200, {
      'Access-Control-Allow-Origin': '*',
      'Cache-Control': 'max-age=0',
      'Content-Length': Buffer.byteLength(content),
      'Content-Type': mime.getType(getTypeFromPath$1(filePath) || filePath)
    });
    res.end(content);

    info$4(
      `${stopwatch.stop(
        res.url,
        true,
        true
      )} handled transpiled request for ${chalk.green(getProjectPath$3(res.url))}`
    );
  }
};

const debug$7 = debug$a('dvlp:watch');
const { FSWatcher } = chokidar;
const { getProjectPath: getProjectPath$4 } = file;


// Hidden, node_modules/*, .dvlp/*
const RE_IGNORED = /(^|[/\\])\..|\.dvlp/i;
const TIMEOUT = 1000;

/**
 * Instantiate a file watcher and begin watching for changes
 *
 * @param { (callback: string) => void } fn
 * @returns { Watcher }
 */
var watch = function watch(fn) {
  const watcher = new FSWatcher({
    ignored: RE_IGNORED,
    ignoreInitial: true,
    persistent: true
  });
  const files = new Set();
  let changing = false;

  watcher.on('unlink', (filePath) => {
    debug$7(`unwatching file "${getProjectPath$4(filePath)}"`);
    watcher.unwatch(filePath);
    files.delete(path.resolve(filePath));
  });
  watcher.on('change', (filePath) => {
    if (!changing) {
      // Prevent double change
      setTimeout(() => {
        changing = false;
      }, TIMEOUT);
      changing = true;
      debug$7(`change detected "${getProjectPath$4(filePath)}"`);
      fn(path.resolve(filePath));
    }
  });

  return {
    add(filePath) {
      if (Array.isArray(filePath)) {
        // @ts-ignore
        return filePath.forEach(this.add);
      }

      filePath = path.resolve(filePath);

      if (!files.has(filePath) && !RE_IGNORED.test(filePath)) {
        debug$7(`watching file "${getProjectPath$4(filePath)}"`);
        files.add(filePath);
        watcher.add(filePath);
      }
    },
    close() {
      watcher.close();
      files.clear();
    }
  };
};

const { error: error$5, fatal: fatal$1, info: info$5, noisyInfo: noisyInfo$3 } = log;
const {
  find: find$3,
  getProjectPath: getProjectPath$5,
  getTypeFromRequest,
  importModule
} = file;
const {
  interceptFileRead: interceptFileRead$1,
  interceptProcessOn: interceptProcessOn$1
} = intercept;
const {
  isModuleBundlerFilePath: isModuleBundlerFilePath$2,
  isNodeModuleFilePath: isNodeModuleFilePath$3
} = is;
const { bundle: bundle$2 } = bundler_1;
const {
  concatScripts: concatScripts$1,
  getDvlpGlobalString: getDvlpGlobalString$1,
  getProcessEnvString: getProcessEnvString$1,
  hashScript: hashScript$1
} = scripts;
const { connectClient: connectClient$1, pushEvent: pushEvent$1 } = pushEvents;



const debug$8 = debug$a('dvlp:server');



// Work around @rollup/plugin-commonjs require.cache
// @ts-ignore
const moduleCache = module$1._cache;

const { patchResponse: patchResponse$1 } = patch;






/** @typedef {import("http").Server} Server */

const START_TIMEOUT_DURATION = 2000;

const { EventSource: EventSource$2 } = fayeWebsocket;
const originalCreateServer = http.createServer;
/** @type { Array<string> } */
let dvlpModules;
/** @type { Array<string> } */
let globalKeys;

var server = class DvlpServer {
  /**
   * Constructor
   *
   * @param { string | (() => void) | undefined } main
   * @param { Reloader } [reloader]
   * @param { object } [rollupConfig]
   * @param { string } [transpilerPath]
   * @param { string | Array<string> } [mockPath]
   */
  constructor(main, reloader, rollupConfig, transpilerPath, mockPath) {
    // Listen for all upcoming file system reads (including require('*'))
    // Register early to catch all reads, including transpilers that patch fs.readFile
    this.watcher = this.createWatcher();
    this.unlistenForFileRead = interceptFileRead$1((filePath) => {
      this.addWatchFiles(filePath);
    });

    if (dvlpModules === undefined) {
      dvlpModules = Object.keys(moduleCache);
    }
    if (globalKeys === undefined) {
      globalKeys = Object.keys(commonjsGlobal);
    }

    /** @type { Array<string> } */
    this.appModules = [];
    this.connections = new Map();
    this.exitHandler = null;
    this.lastChanged = '';
    this.main = main;
    this.mocks = mockPath ? new mock_1(mockPath) : undefined;

    const headerScript = concatScripts$1([
      getProcessEnvString$1(),
      getDvlpGlobalString$1(),
      (this.mocks && this.mocks.client) || ''
    ]);

    this.patchResponseOptions = {
      rollupConfig,
      footerScript: {
        hash: reloader && hashScript$1(reloader.client),
        string: reloader ? reloader.client : '',
        url: reloader && reloader.url
      },
      headerScript: {
        hash: hashScript$1(headerScript),
        string: headerScript
      }
    };
    this.origin = '';
    this.port = Number(process.env.PORT);
    this.reloader = reloader;
    this.rollupConfig = rollupConfig;
    /** @type { Server | null } */
    this.server = null;
    this.transpiler = transpilerPath && importModule(transpilerPath);
    this.transpilerCache = this.transpiler && new Map();
    this.urlToFilePath = new Map();
  }

  /**
   * Create watcher instance and react to file changes
   *
   * @returns { Watcher }
   */
  createWatcher() {
    return watch(async (filePath) => {
      noisyInfo$3(
        `\n  ⏱  ${new Date().toLocaleTimeString()} ${chalk.yellow(
          getProjectPath$5(filePath)
        )}`
      );

      this.lastChanged = filePath;

      try {
        // TODO: conditional upon client-only file?
        await this.restart();

        for (const [url, fp] of this.urlToFilePath) {
          if (fp === filePath) {
            filePath = url;
            break;
          }
        }

        // Will trigger one of:
        // 1. single css refresh if css and a link.href matches filePath
        // 2. multiple css refreshes if css and no link.href matches filePath (ie. it's a dependency)
        // 3. full page reload
        this.reloader && this.reloader.send(filePath);
      } catch (err) {
        error$5(err);
      }
    });
  }

  /**
   * Add "filePaths" to watcher
   *
   * @param { string | Array<string> } filePaths
   */
  addWatchFiles(filePaths) {
    if (!Array.isArray(filePaths)) {
      filePaths = [filePaths];
    }

    for (const filePath of filePaths) {
      if (!isNodeModuleFilePath$3(filePath)) {
        this.watcher.add(filePath);
      }
    }
  }

  /**
   * Start app server
   * Proxies createServer to grab instance and register for events
   *
   * @returns { Promise<void> }
   */
  start() {
    return new Promise((resolve, reject) => {
      // Force reject if never started
      const timeoutID = setTimeout(() => {
        debug$8('server not started after timeout');
        reject(Error('unable to start server'));
      }, START_TIMEOUT_DURATION);
      const instance = this;

      http.createServer = new Proxy(http.createServer, {
        apply(target, ctx, args) {
          // Handler always last arg ('options' as first arg added in 9.6)
          let handler = args[args.length - 1];

          // Wrap request handler (if passed)
          if (handler && typeof handler === 'function') {
            args[args.length - 1] = instance.createRequestHandler(
              instance,
              handler
            );
          } else {
            handler = undefined;
          }

          /** @type { Server } */
          const server = (instance.server = Reflect.apply(target, ctx, args));

          server.timeout = server.keepAliveTimeout = 0;

          server.on('connection', (connection) => {
            const key = `${connection.remoteAddress}:${connection.remotePort}`;

            instance.connections.set(key, connection);
            connection.on('close', () => {
              instance.connections.delete(key);
            });
          });
          server.on('error', (err) => {
            instance.stop();
            reject(err);
          });
          server.on('listening', () => {
            debug$8('server started');
            clearTimeout(timeoutID);
            instance.appModules = getAppModules();
            instance.addWatchFiles(instance.appModules);
            const address = server.address();
            if (address && typeof address !== 'string') {
              instance.port = address.port;
            }
            instance.origin = `http://localhost:${instance.port}`;
            resolve();
          });
          server.on('upgrade', (req, socket, body) => {
            handleMockWebSocket(req, socket, body, instance.mocks);
          });

          // No handler registered so proxy "request" listener
          if (!handler) {
            server.on = new Proxy(server.on, {
              apply(target, ctx, args) {
                const [eventName, listener] = args;

                // Wrap request handler
                if (eventName === 'request') {
                  args[1] = instance.createRequestHandler(instance, listener);
                }

                Reflect.apply(target, ctx, args);
              }
            });
          }

          // Un-proxy in case more than one server created
          // (assumes first server is application server)
          http.createServer = originalCreateServer;

          return server;
        }
      });

      try {
        // Trigger application bootstrap
        this.startApplication();
      } catch (err) {
        reject(err);
      }
    });
  }

  /**
   * Create request handler wrapper for 'originalRequestHandler'
   *
   * @param { DvlpServer } server
   * @param { RequestHandler } originalRequestHandler
   * @returns { RequestHandler }
   */
  createRequestHandler(server, originalRequestHandler) {
    return async function requestHandler(req, res) {
      stopwatch.start(req.url);

      const type = getTypeFromRequest(req);
      let filePath = server.urlToFilePath.get(req.url);

      res.url = req.url;

      if (
        handleMockResponse(req, res, server.mocks) ||
        handlePushEvent(req, res, server.mocks)
      ) {
        return;
      }

      // Ignore html or uncached or no longer available at previously known path
      if (type !== 'html' && (!filePath || !fs.existsSync(filePath))) {
        filePath = find$3(req, { type });

        if (filePath) {
          server.urlToFilePath.set(req.url, filePath);

          if (isModuleBundlerFilePath$2(filePath)) {
            await bundle$2(
              path.basename(filePath),
              undefined,
              undefined,
              server.rollupConfig
            );
          }
        } else {
          // File not found. Clear previously known path
          server.urlToFilePath.delete(req.url);
        }
      }

      // Ignore unknow types
      if (type) {
        patchResponse$1(filePath, req, res, server.patchResponseOptions);
      }

      if (filePath) {
        const bundled = isModuleBundlerFilePath$2(filePath);

        // node_modules file (bundled or native esm)
        if (bundled || isNodeModuleFilePath$3(filePath)) {
          info$5(
            `${stopwatch.stop(res.url, true, true)} handled${
              bundled ? ' bundled' : ''
            } request for ${chalk.green(getProjectPath$5(req.url))}`
          );

          return send(req, filePath, {
            cacheControl: true,
            dotfiles: 'allow',
            maxAge: config.maxAge
          }).pipe(res);
        } else if (server.transpiler) {
          // Will respond if transpiler for this type
          await transpile(filePath, res, {
            transpilerCache: server.transpilerCache,
            lastChanged: server.lastChanged,
            transpiler: server.transpiler
          });
        }
      }

      // Not mocked/transpiled/bundled, so pass through request to app
      if (!res.finished) {
        debug$8(`allowing app to handle "${req.url}"`);
        originalRequestHandler(req, res);
      }
    };
  }

  /**
   * Start application
   *
   * @returns { void }
   */
  startApplication() {
    debug$8('starting application');
    process.on('uncaughtException', this.onUncaught);
    // @ts-ignore
    process.on('unhandledRejection', this.onUncaught);
    // Listen for process event registration to capture any exit handlers
    this.unlistenForProcessOn = interceptProcessOn$1((event, handler) => {
      if (event === 'exit' || event === 'beforeExit') {
        this.exitHandler = handler;
        return false;
      }
    });
    if (!this.main) {
      fileServer();
    } else if (typeof this.main === 'function') {
      this.main();
    } else {
      importModule(this.main, this.transpiler);
    }
  }

  /**
   * Stop application
   *
   * @returns { Promise<void> }
   */
  async stopApplication() {
    debug$8('stopping application');
    if (this.unlistenForProcessOn) {
      this.unlistenForProcessOn();
    }
    if (this.exitHandler) {
      await this.exitHandler();
    }
    process.removeListener('uncaughtException', this.onUncaught);
    process.removeListener('unhandledRejection', this.onUncaught);
    // @ts-ignore
    clearAppModules(this.appModules, this.main);
    clearGlobals();
  }

  /**
   * Stop running server
   *
   * @returns { Promise<void> }
   */
  stop() {
    return new Promise(async (resolve) => {
      for (const connection of this.connections.values()) {
        connection.destroy();
      }
      this.connections.clear();

      await this.stopApplication();

      if (!this.server) {
        return resolve();
      }

      debug$8('server stopped');
      this.server.removeAllListeners();
      if (!this.server.listening) {
        resolve();
      } else {
        this.server.close(() => {
          resolve();
        });
      }
    });
  }

  /**
   * Restart running server
   *
   * @returns { Promise<void> }
   */
  async restart() {
    if (this.main) {
      debug$8('server restarting');
      await this.stop();
      return this.start();
    }
  }

  /**
   * Handle uncaughtException & unhandledRejection
   * Logs errors to console
   *
   * @param { Error } err
   * @returns { void }
   */
  onUncaught(err) {
    fatal$1(err);
  }

  /**
   * Destroy instance
   *
   * @returns { Promise<void> }
   */
  destroy() {
    debug$8('server destroying');
    this.mocks && this.mocks.clean();
    this.unlistenForFileRead();
    this.watcher.close();
    return this.stop();
  }
};

/**
 * Handle mock responses, including EventSource connection
 * Returns 'true' if handled
 *
 * @param { Req } req
 * @param { Res } res
 * @param { Mock } [mocks]
 * @returns { boolean }
 */
function handleMockResponse(req, res, mocks) {
  const url = new URL(req.url, `http://localhost:${config.activePort}`);
  let mock = url.searchParams.get('dvlpmock');

  if (mocks && mock) {
    mock = decodeURIComponent(mock);

    // @ts-ignore
    if (EventSource$2.isEventSource(req)) {
      connectClient$1(
        {
          url: mock,
          type: 'es'
        },
        req,
        res
      );
      // Send 'connect' event if it exists
      mocks.matchPushEvent(mock, 'connect', pushEvent$1);
      noisyInfo$3(
        `${stopwatch.stop(
          req.url,
          true,
          true
        )} connected to EventSource client at ${chalk.green(mock)}`
      );
    } else {
      mocks.matchResponse(mock, req, res);
    }

    return true;
  }

  return false;
}

/**
 * Handle mock WebSocket connection
 *
 * @param { Req } req
 * @param { object } socket
 * @param { object } body
 * @param { Mock } [mocks]
 * @returns { void }
 */
function handleMockWebSocket(req, socket, body, mocks) {
  const url = new URL(req.url, `http://localhost:${config.activePort}`);
  let mock = url.searchParams.get('dvlpmock');

  if (mocks && mock && fayeWebsocket.isWebSocket(req)) {
    mock = decodeURIComponent(mock);
    connectClient$1(
      {
        url: mock,
        type: 'ws'
      },
      req,
      socket,
      body
    );
    // Send 'connect' event if it exists
    mocks.matchPushEvent(mock, 'connect', pushEvent$1);
    noisyInfo$3(
      `${stopwatch.stop(
        req.url,
        true,
        true
      )} connected to WebSocket client at ${chalk.green(mock)}`
    );
  }
}

/**
 * Handle push event request
 * Returns 'true' if handled
 *
 * @param { Req } req
 * @param { Res } res
 * @param { Mock } [mocks]
 * @returns { boolean }
 */
function handlePushEvent(req, res, mocks) {
  if (mocks && req.method === 'POST' && req.url === '/dvlp/push-event') {
    let body = '';

    req.on('data', (chunk) => {
      body += chunk.toString();
    });
    req.on('end', () => {
      const { stream, event } = JSON.parse(body);

      if (typeof event === 'string') {
        mocks.matchPushEvent(stream, event, pushEvent$1);
      } else {
        pushEvent$1(stream, event);
      }

      res.writeHead(200);
      res.end('ok');
    });

    return true;
  }

  return false;
}

/**
 * Retrieve app modules (excluding node_modules)
 *
 * @returns { Array<string> }
 */
function getAppModules() {
  const modules = Object.keys(moduleCache).filter(
    (m) => !dvlpModules.includes(m) /* && !isNodeModuleFilePath(m) */
  );

  debug$8(`found ${modules.length} app modules`);

  return modules;
}

/**
 * Clear app modules from module cache
 *
 * @param { Array<string> } appModules
 * @param { string } main
 */
function clearAppModules(appModules, main) {
  const mainModule = moduleCache[main];

  // Remove main from parent
  // (No children when bundled)
  if (
    mainModule !== undefined &&
    mainModule.parent !== undefined &&
    mainModule.parent.children !== undefined
  ) {
    const parent = mainModule.parent;
    let i = parent.children.length;

    while (--i) {
      if (parent.children[i].id === mainModule.id) {
        parent.children.splice(i, 1);
      }
    }
  }

  for (const m of appModules) {
    delete moduleCache[m];
  }

  debug$8(`cleared ${appModules.length} app modules from require.cache`);
}

/**
 * Clear any added globals
 */
function clearGlobals() {
  for (const key of Object.keys(commonjsGlobal)) {
    if (!globalKeys.includes(key)) {
      // @ts-ignore
      commonjsGlobal[key] = undefined;
    }
  }
}

const {
  exists,
  expandPath,
  getProjectPath: getProjectPath$6,
  importModule: importModule$1
} = file;
const { info: info$6, error: error$6 } = log;


const { destroyWorkers: destroyWorkers$1 } = bundler_1;





/**
 * Server instance factory
 *
 * @param { string | Array<string> | (() => void) } filePath
 * @param { ServerOptions } options
 * @returns { Promise<Server> }
 */
var server$1 = async function serverFactory(
  filePath = process.cwd(),
  { mockPath, port = config.port, reload = true, silent, transpiler } = {}
) {
  const isFunction = typeof filePath === 'function';
  let isStatic = false;

  if (!isFunction) {
    // @ts-ignore
    filePath = expandPath(filePath);

    if (Array.isArray(filePath) && filePath.length === 1) {
      filePath = filePath[0];
    }

    exists(filePath);

    isStatic =
      Array.isArray(filePath) ||
      fs.statSync(path.resolve(filePath)).isDirectory();
    /** @type { Array<string> } */
    const directories = [];

    for (let directory of [
      process.cwd(),
      ...(Array.isArray(filePath) ? filePath : [filePath])
    ]) {
      directory = path.resolve(directory);
      if (fs.statSync(directory).isFile()) {
        directory = path.dirname(directory);
      }

      const nodeModules = path.join(directory, 'node_modules');

      directories.push(directory);
      if (fs.existsSync(nodeModules)) {
        directories.push(nodeModules);
      }
    }

    config.directories = Array.from(new Set(directories));
  }

  if (mockPath) {
    mockPath = expandPath(mockPath);
  }

  /** @type { Reloader | undefined } */
  let reloader$1;
  let rollupConfig;
  let transpilerPath;

  if (process.env.PORT === undefined) {
    process.env.PORT = String(port);
  }

  if (fs.existsSync(config.rollupConfigPath)) {
    rollupConfig = importModule$1(config.rollupConfigPath);
    info$6(
      `${chalk.green('✔')} registered custom Rollup.js config at ${chalk.green(
        getProjectPath$6(config.rollupConfigPath)
      )}`
    );
  }
  if (transpiler) {
    transpilerPath = path.resolve(transpiler);

    info$6(
      `${chalk.green('✔')} loaded transpiler from ${chalk.green(
        getProjectPath$6(transpilerPath)
      )}`
    );
  }
  if (reload) {
    reloader$1 = await reloader();
  }

  const main = isStatic
    ? undefined
    : isFunction
    ? filePath
    : path.resolve(filePath);

  const server$1 = new server(
    main,
    reloader$1,
    rollupConfig,
    transpilerPath,
    mockPath
  );

  try {
    await server$1.start();
  } catch (err) {
    error$6(err);
  }

  const parentDir = path.resolve(process.cwd(), '..');
  const paths = isStatic
    ? config.directories
        .map((dir) => path.relative(parentDir, dir) || path.basename(parentDir))
        .join(', ')
    : isFunction
    ? 'function'
    : getProjectPath$6(filePath);

  info$6(
    `\n  💥 serving ${chalk.green(paths)} at ${chalk.green.underline(
      server$1.origin
    )}`
  );

  config.activePort = server$1.port;

  info$6(' 👀 watching for changes...\n');

  if (silent) {
    log.silent = true;
  }

  return {
    port: server$1.port,
    restart: server$1.restart.bind(server$1),
    destroy() {
      return Promise.all([
        reloader$1 && reloader$1.destroy(),
        server$1.destroy(),
        destroyWorkers$1()
      ]).then();
    }
  };
};

const {
  connectClient: connectClient$2,
  destroyClients: destroyClients$1,
  pushEvent: pushEvent$2
} = pushEvents;

const debug$9 = debug$a('dvlp:test');



const { interceptClientRequest: interceptClientRequest$2 } = intercept;
const { isLocalhost: isLocalhost$2 } = is;



const { URL: URL$3 } = url$1;


const { EventSource: EventSource$3 } = fayeWebsocket;
/** @type { Set<TestServerInstance> } */
const instances = new Set();
let reroute = false;
let networkDisabled = false;
/** @type { () => void } */
let uninterceptClientRequest;

/**
 * Create test server
 *
 * @param { TestServerOptions } [options]
 * @returns { Promise<TestServerInstance> }
 */
var testServer = async function testServerFactory(options) {
  enableRequestIntercept();

  /** @type { TestServerInstance } */
  const server = new TestServer(options || {});

  // @ts-ignore: private
  await server._start();
  // Make sure 'mock' has access to current active port
  config.activePort = server.port;

  instances.add(server);

  return server;
};

/**
 * Disable all external network connections
 * and optionally reroute all external requests to this server
 *
 * @param { boolean } [rerouteAllRequests]
 * @returns { void }
 */
var disableNetwork = function disableNetwork(
  rerouteAllRequests = false
) {
  enableRequestIntercept();
  networkDisabled = true;
  reroute = rerouteAllRequests;
};

/**
 * Re-enable all external network connections
 *
 * @returns { void }
 */
var enableNetwork = function enableNetwork() {
  enableRequestIntercept();
  networkDisabled = true;
  reroute = false;
};

/**
 * Default mock response handler for network hang
 *
 * @param { Req } req
 * @param { Res } res
 * @returns { undefined }
 */
var mockHangResponseHandler = function mockHangResponseHandler(
  req,
  res
) {
  return;
};

/**
 * Default mock response handler for 500 response
 *
 * @param { Req } req
 * @param { Res } res
 * @returns { undefined }
 */
var mockErrorResponseHandler = function mockErrorResponseHandler(
  req,
  res
) {
  res.writeHead(500);
  res.error = Error('error');
  res.end('error');
  return;
};

/**
 * Default mock response handler for 404 response
 *
 * @param { Req } req
 * @param { Res } res
 * @returns { undefined }
 */
var mockMissingResponseHandler = function mockMissingResponseHandler(
  req,
  res
) {
  res.writeHead(404);
  res.end('missing');
  return;
};

/**
 * Default mock response handler for offline
 *
 * @param { Req } req
 * @param { Res } res
 * @returns { undefined }
 */
var mockOfflineResponseHandler = function mockOfflineResponseHandler(
  req,
  res
) {
  req.socket.destroy();
  return;
};

class TestServer {
  /**
   * Constructor
   *
   * @implements { TestServerInstance }
   * @param { TestServerOptions } options
   */
  constructor(options) {
    const {
      autorespond = true,
      latency = config.latency,
      port = config.port,
      webroot = ''
    } = options;

    this.latency = latency;
    /** @type { MockInstance } */
    this.mocks = new mock_1();
    this.webroot = webroot;
    this._autorespond = autorespond;
    this.port = port;
    this._server = null;
  }

  /**
   * Start server
   *
   * @returns { Promise<void> }
   * @private
   */
  _start() {
    return new Promise((resolve, reject) => {
      this._server = http.createServer(async (req, res) => {
        // @ts-ignore
        if (EventSource$3.isEventSource(req)) {
          connectClient$2(
            // @ts-ignore
            {
              url: req.url,
              type: 'es'
            },
            req,
            res
          );
          // @ts-ignore
          this.pushEvent(req.url, 'connect');
          return;
        }

        // @ts-ignore
        const url = new URL$3(req.url, `http://localhost:${this.port}`);
        const error = url.searchParams.get('error') != null;
        const hang = url.searchParams.get('hang') != null;
        const maxage = url.searchParams.get('maxage') || 0;
        const missing = url.searchParams.get('missing') != null;
        const offline = url.searchParams.get('offline') != null;
        const mock = url.searchParams.get('dvlpmock');

        if (hang) {
          return;
        }

        this.latency && (await sleep$1(this.latency));

        if (error || missing) {
          const statusCode = error ? 500 : 404;
          const body = error ? 'error' : 'missing';

          debug$9(`not ok: ${req.url} responding with ${statusCode}`);
          res.statusCode = statusCode;
          res.end(body);
          return;
        } else if (offline) {
          debug$9(`not ok: ${req.url} offline`);
          req.socket.destroy();
          return;
          // @ts-ignore
        } else if (req.aborted) {
          debug$9(`not ok: ${req.url} aborted`);
          return;
        } else if (mock) {
          debug$9(`ok: ${req.url} responding with mocked data`);
          // @ts-ignore
          this.mocks.matchResponse(mock, req, res);
          return;
        }

        const trimmedPath = url.pathname.slice(1);
        let type = mime.getType(trimmedPath);
        // TODO: handle encoded query strings in path name?
        let filePath = path.resolve(path.join(this.webroot, trimmedPath));
        let body = '';
        let headers = {};
        let size = 0;
        let stat;
        let msg = '';

        // Ignore webroot if no file
        if (!fs.existsSync(filePath)) {
          filePath = path.resolve(trimmedPath);
        }

        try {
          stat = fs.statSync(filePath);
          size = stat.size;
          msg = `ok: ${req.url} responding with file`;
        } catch (err) {
          if (!this._autorespond) {
            res.writeHead(404);
            return res.end();
          }
          body = '"hello"';
          size = Buffer.byteLength(body);
          msg = `ok: ${req.url} responding with dummy file`;
        }

        // @ts-ignore
        res.writeHead(200, {
          'Content-Length': size,
          'Cache-Control': `public, max-age=${maxage}`,
          'Content-Type': type,
          ...headers
        });

        debug$9(msg);

        return body ? res.end(body) : fs.createReadStream(filePath).pipe(res);
      });

      serverDestroy(this._server);
      this._server.unref();
      this._server.on('error', reject);
      this._server.on('listening', resolve);
      this._server.on('upgrade', (req, socket, body) => {
        if (fayeWebsocket.isWebSocket(req)) {
          connectClient$2(
            {
              url: req.url,
              type: 'ws'
            },
            req,
            socket,
            body
          );

          this.pushEvent(
            new URL$3(req.url, `ws://${req.headers.host}`).href,
            'connect'
          );
        }
      });

      this._server.listen(this.port);
    });
  }

  /**
   * Load mock files at 'filePath'
   *
   * @param { string | Array<string> } filePath
   */
  loadMockFiles(filePath) {
    this.mocks.load(filePath);
  }

  /**
   * Register mock 'response' for 'request'
   *
   * @param { string | MockRequest } request
   * @param { MockResponse | MockResponseHandler } response
   * @param { boolean } [once]
   * @param { () => void } [onMockCallback]
   * @returns { void }
   */
  mockResponse(request, response, once = false, onMockCallback) {
    this.mocks.addResponse(request, response, once, onMockCallback);
  }

  /**
   * Register mock push 'events' for 'stream'
   *
   * @param { string | MockPushStream } stream
   * @param { MockPushEvent | Array<MockPushEvent> } events
   * @returns { void }
   */
  mockPushEvents(stream, events) {
    this.mocks.addPushEvents(stream, events);
  }

  /**
   * Push data to WebSocket/EventSource clients
   * A string passed as 'event' will be handled as a named mock push event
   *
   * @param { string | PushStream } stream
   * @param { string | PushEvent } [event]
   * @returns { void }
   */
  pushEvent(stream, event) {
    // Passed a mocked event name
    if (typeof event === 'string') {
      this.mocks.matchPushEvent(stream, event, pushEvent$2);
    } else {
      // @ts-ignore
      pushEvent$2(stream, event);
    }
  }

  /**
   * Stop running server
   *
   * @returns { Promise<void> }
   * @private
   */
  _stop() {
    return new Promise((resolve) => {
      if (!this._server) {
        return resolve();
      }

      this._server.removeAllListeners();
      // @ts-ignore
      this._server.destroy(() => {
        debug$9('server stopped');
        resolve();
      });
    });
  }

  /**
   * Destroy instance
   *
   * @returns { Promise<void> }
   */
  destroy() {
    debug$9('destroying');
    destroyClients$1();
    this.mocks.clean();
    instances.delete(this);
    return this._stop();
  }
}

/**
 * Enable request interception to allow mocking/network disabling
 */
function enableRequestIntercept() {
  if (uninterceptClientRequest === undefined) {
    uninterceptClientRequest = interceptClientRequest$2((url) => {
      const isMocked = Array.from(instances).some((instance) =>
        instance.mocks.hasMatch(url)
      );
      let hostname = url.hostname || url.host;

      // Allow mocked requests to pass-through
      if (!isMocked && !isLocalhost$2(hostname)) {
        if (reroute) {
          // Reroute back to this server
          url.host = url.hostname = `localhost:${config.activePort}`;
        } else if (networkDisabled) {
          throw Error(`network connections disabled. Unable to request ${url}`);
        }
      }

      return true;
    });
  }
}

/**
 * Sleep for random number of milliseconds between 'min' and '2xmin'
 *
 * @param { number } min
 * @returns { Promise<void> }
 * @private
 */
function sleep$1(min) {
  return new Promise((resolve) => {
    if (!min) {
      return resolve();
    }
    setTimeout(resolve, min + Math.random() * min);
  });
}
testServer.disableNetwork = disableNetwork;
testServer.enableNetwork = enableNetwork;
testServer.mockHangResponseHandler = mockHangResponseHandler;
testServer.mockErrorResponseHandler = mockErrorResponseHandler;
testServer.mockMissingResponseHandler = mockMissingResponseHandler;
testServer.mockOfflineResponseHandler = mockOfflineResponseHandler;

/** @type { server } */
var server$2 = server$1;
/** @type { testServer } */
var testServer$1 = testServer;

var lib = {
	server: server$2,
	testServer: testServer$1
};

exports.default = lib;
exports.server = server$2;
exports.testServer = testServer$1;
